from utilities.intspan import subsume, overlap, subtract, cardinality
from utilities.tools import flatten, reverse_complement, is_integer
from utilities.bam import BAM
import re
from pep_change import translate, pep_change

class Event:

    splice_motifs = False
    
    headers = [
        'type',
        'contig',
        'transcript',
        'gene',
        'exons',
        'align_blocks',
        'coord',
        'splice',
        'multi_3',
        'size',
        'orf'
        ]

    headers_support = [
        'spanning_reads',
        'contig_coverage',
        'contig_neighbor',
        'read_support',
        'filter'
        ]

    def __init__(self, type):
        self.type = type

        self.contig = None
        self.transcript = None
        self.model = None
        self.exons = []
        self.chrom = None
        self.exon_coords = []
        self.align_blocks = []
        self.align_coords = []
        self.novel = True
        self.artefact = False
        #self.match_result = False
        self.align = None
        self.txt = None
        self.filter = None
        self.orf = None


    def details(self, refseq=None):
        '''
        sets other attributes of events
        '''
        
        if self.type == "retained_intron":
            apart = int(self.exon_coords[1][0]) - 1 - int(self.exon_coords[0][1])
            self.multi_3 = self.multiple_3(apart)
        else:
            self.multi_3 = None

        #splice
        self.splice = None
        if self.type in ("AS5", "AS3"):
            ss = self.get_splice_sites(self.align_blocks[0], left=self.left, right=self.right)
            if ss:
                self.splice = ss[0]
        elif self.type in ("novel_exon", "AS53", "novel_transcript", "novel_utr", "novel_intron"):
            splice_sites = []
            for block in self.align_blocks:
                ss = ",".join(self.get_splice_sites(block, left=True, right=True))
                if ss:
                    splice_sites.append(ss)
            if splice_sites:
                self.splice = ";".join(splice_sites)

        #size
        self.size = None
        if self.type in ("novel_exon", "AS53", "novel_transcript", "novel_utr", "novel_intron"):
            sizes = []
            for block in self.align_coords:
                sizes.append(str(cardinality(block)))

            self.size = ",".join(sizes)

        #orf
        if refseq:
            self.find_pep_change(refseq)
            

    def output(self, gff=False):
        data = {}

        #initialize
        for header in Event.headers:
            data[header] = 'na'
            
        data['type'] = self.type
        data['contig'] = self.contig
        data['align_blocks'] = ",".join([str(b) for b in self.align_blocks])
        data['coord'] = self.coordinate()


        if self.type != 'novel_transcript':
            data['transcript'] = self.txt.name
            if self.txt.alias:
                data['gene'] = self.txt.alias
            else:
                data['gene'] = 'na'
                
            exons = []
            for e in self.exons:
                exons.append(self.convert_gene_exon_number(self.txt, int(e)))
            data['exons'] = ",".join([str(e) for e in exons])

        for item in ('size', 'multi_3', 'splice', 'orf'):
            try:
                value = getattr(self, item)
            except AttributeError:
                pass
            else:
                if value:
                    data[item] = value

        if self.filter:
            for item in Event.headers_support:
                try:
                    value = getattr(self, item)

                except AttributeError:
                    data[item] = 'na'
                    
                else:
                    if value:
                        if item == 'contig_neighbor' and self.type != 'novel_transcript':
                            value = ','.join([str(value['left']), str(value['right'])])
                            
                        elif value != 'na' and type(value).__name__ == 'list':
                            value = [str(v) for v in value]
                            value = ','.join(value)
                            
                        elif item == 'read_support' and value != 'na':
                            if value:
                                value = 'passed'
                            else:
                                value = 'failed'
                            
                        data[item] = str(value)
                    else:
                        data[item] = 'na'

        fields = []
        for item in Event.headers:
            fields.append(data[item])
        if self.filter:
            for item in Event.headers_support:
                fields.append(data[item])

        if gff:
            gff_id, start, end = re.split('[:-]', data['coord'])
            gff_type = 'exon'
            if self.type in ("AS5", "AS3", 'AS53'):
                gff_type = 'splice_site'
            elif self.type == 'novel_intron':
                gff_type = 'intron'
            elif self.type == 'novel_utr':
                gff_type = 'UTR'
            elif self.type == 'novel_transcript':
                gff_type = 'transcript_region'

            score = '.'
            strand = '+'
            phase = '.'
            attributes_list = []
            for i in range(len(Event.headers)):
                attributes_list.append(Event.headers[i] + '=' + fields[i])
            attributes_str = ';'.join(attributes_list)

            source = 'trans-ABySS'
            return ('\t'.join((gff_id, source, gff_type, start, end, score, strand, phase, attributes_str)))
            
        else:
            return ('\t'.join(fields))
        
    def convert_gene_exon_number(self, txt, positive_exon_num):
        '''
        convert exon number to one taking gene strand into consideration
        positive_exon_num assumed 1-based
        '''
        exon_num = positive_exon_num
        if txt.strand == '-':
            exon_num = len(txt.exons) - positive_exon_num + 1
            
        return exon_num

    def get_splice_sites(self, block, left=False, right=False):
        ss = []
        if not Event.splice_motifs:
            return ss
        
        align = self.align
       
        if align and int(block) >= 0 and align.splice_sites and (left or right):
            splice_sites = align.splice_sites

            if splice_sites[0] == '':
                splice_sites.pop(0)

            if left and int(block) > 1:
                splice_site = splice_sites[block-2]
                #print "left", block, splice_site
                if align.orient and align.orient == '-':
                    splice_site = reverse_complement(splice_site)
                #print "left", block, splice_site, align.orient

                motif = "?"
                if Event.splice_motifs.has_key(splice_site):
                    motif = Event.splice_motifs[splice_site]

                donor = splice_site[:2]
                acceptor = splice_site[-2:]

                if align.orient and align.orient == '-':
                    donor = donor.upper()
                    acceptor = acceptor.lower()
                else:
                    donor = donor.lower()
                    acceptor = acceptor.upper()

                if donor and acceptor:
                    ss.append("-".join([donor, acceptor]) + "(%s)" % (motif))

                #print "left:",block,donor,acceptor

            if right and int(block) < len(align.blocks):
                splice_site = splice_sites[block-1]
                #print "right", block, splice_site

                if align.orient and align.orient == '-':
                    splice_site = reverse_complement(splice_site)
                #print "right", block, splice_site, align.orient

                motif = "?"
                if Event.splice_motifs.has_key(splice_site):
                    motif = Event.splice_motifs[splice_site]

                donor = splice_site[:2]
                acceptor = splice_site[-2:]

                if align.orient and align.orient == '-':
                    donor = donor.lower()
                    acceptor = acceptor.upper()
                else:
                    donor = donor.upper()
                    acceptor = acceptor.lower()

                #print "right:",block,donor,acceptor

                if donor and acceptor:
                    ss.append("-".join([donor, acceptor]) + "(%s)" % (motif))

        return ss

                    
    def coordinate(self):
        coord = ""

        if self.type == "skipped_exon" or self.type == "retained_intron":
            coord = self.chrom + ":" + str(self.exon_coords[0][0]) + "-" + str(self.exon_coords[-1][1])
        elif self.type == "novel_intron":
            coord = self.chrom + ":" + str(int(self.align_coords[0][1])+1) + "-" + str(int(self.align_coords[-1][0])-1)
        elif self.type == "novel_exon" or self.type == "novel_utr" or self.type == "novel_transcript" or self.type == "AS53":
            coord = self.chrom + ":" + str(self.align_coords[0][0]) + "-" + str(self.align_coords[-1][1])
        elif self.type == "AS5" or self.type == "AS3":
            coord = self.chrom + ":" + str(self.align_coords[0]) + "-" + str(self.align_coords[-1])

        return coord

    def multiple_3(self, apart):
        if apart % 3 == 0:
            return 'True'
        else:
            return 'False'

    def set_novelty(self, txts, matches=None):
        novel_events = []
        
        if self.type == "novel_exon" or self.type == "AS53" or self.type == "novel_utr":
        
            for txt in txts:
                blocks_to_delete = []

                for b in range(len(self.align_blocks)):
                    for e in range(len(txt.exons)):
                        novel = True

                        if subsume(self.align_coords[b], txt.exons[e]):
                            novel = False

                        #novel utr - requires just one edge to align
                        if novel and self.type == 'novel_utr':
                            if int(self.align_coords[b][0]) == int(txt.exons[e][0]) or int(self.align_coords[b][1]) == int(txt.exons[e][1]):
                                novel = False
                                
                        if not novel:
                            blocks_to_delete.append(b)
                            break

                if blocks_to_delete:
                    for b in sorted(blocks_to_delete, reverse=True):
                        del self.align_blocks[b]
                        del self.align_coords[b]

            if not self.align_blocks:
                self.novel = False

        elif self.type == "retained_intron":
            multi = False
            if int(self.exons[-1]) - int(self.exons[0]) > 1:
                multi = True
                self.novel = False

            for i in range(len(self.exon_coords)-1):
                retained_intron = [int(self.exon_coords[i][1])+1, int(self.exon_coords[i+1][0])-1]
                middle_exons = {}
                for txt in txts:
                    exons_txt = []
                    for j in range(len(txt.exons)):
                        exon = txt.exons[j]

                        #terminal exon, require subsume
                        if j == 0 or j == len(txt.exons)-1:
                            if subsume(retained_intron, exon):
                                exons_txt.append(exon) 
                        #middle exons, require just overlap
                        elif overlap(exon, (retained_intron[0], retained_intron[1])):
                            exons_txt.append(exon)

                    if exons_txt:
                        middle_exons[txt] = exons_txt

                #only time when original event is novel WITHOUT testing is when it's a single ri and it is clear of overlapping exons
                if len(middle_exons.keys()) == 0 and not multi:
                    self.novel = True
                
                else:
                    self.novel = False

                    #t = [a.name for a in middle_exons.keys()]

                    #substract overlapping exons
                    if middle_exons.values() and middle_exons.values()[0]:
                        true_retained_intron = subtract(retained_intron, middle_exons.values())
                    else:
                        true_retained_intron = [retained_intron]

                    #if there is still some intron left after subtraction
                    if true_retained_intron:
                        #create new events
                        for ri in true_retained_intron:
                            event = {'contig': self.contig, 'chrom':self.chrom, 'align_blocks':self.align_blocks, 'align_coords':self.align_coords, 'type':self.type, 'novel':True}
                            #print ri, self.transcript, self.novel
                            #determine flanking exons, and transcript by frequency of flanking coordinates
                            flanks = {}
                            for txt in txts:
                                for i in range(len(txt.exons)-1):
                                    left = txt.exons[i]
                                    right = txt.exons[i+1]
                                    #print "ri flanks", txt.full_name(), left, right, ri
                                    if left[1]+1 == ri[0] and right[0]-1 == ri[1]:
                                        #print "ri flanks", txt.full_name(), left, right, ri
                                        coord = ",".join((str(left[0]), str(left[1]), str(right[0]), str(right[1])))
                                        if not flanks.has_key(coord):
                                            flanks[coord] = [[txt,i,i+1]]
                                        else:
                                            flanks[coord].append([txt,i,i+1])

                            #print "flanks", flanks

                            if len(flanks.keys()) > 0:
                                #use the most commom frequent exons
                                flanks_sorted = flanks.keys()
                                flanks_sorted.sort(lambda x,y: len(flanks[y])-len(flanks[x]))

                                #use the orginally assigned transcript if possible
                                same_txt = False
                                exon_coords = flanks_sorted[0].split(',')
                                exons = []
                                for txt,e1,e2 in flanks[flanks_sorted[0]]:
                                    if txt.full_name() == self.transcript:
                                        same_txt = True
                                        event['transcript'] = txt.full_name()
                                        event['exons'] = [e1+1,e2+1]
                                        event['exon_coords'] = [exon_coords[:2], exon_coords[2:]]
                                        event['txt'] = txt
                                        break

                                if not same_txt:
                                    txt,e1,e2 = flanks[flanks_sorted[0]][0]
     
                                    #print "ri new", txt.full_name(), txt.exons, exons, exon_coords
                                    event['transcript'] = txt.full_name()
                                    event['exons'] = [e1+1,e2+1]
                                    event['exon_coords'] = [exon_coords[:2], exon_coords[2:]]
                                    event['txt'] = txt

                                novel_events.append(event)
            

        elif self.type == "skipped_exon":
            for txt in txts:
                for e in range(len(txt.exons)-1):
                    intron_span = [int(txt.exons[e][1])+1, int(txt.exons[e+1][0])-1]

                    if self.align_coords and self.exon_coords:
                        if subsume([int(self.exon_coords[0][0]), int(self.exon_coords[-1][1])], intron_span):
                            self.novel = False
                            break
                        
                if not self.novel:
                    break

        elif self.type == "novel_intron":
            novel_intron_span = [int(self.align_coords[0][1])+1, int(self.align_coords[1][0])-1]
            novel_intron_size = int(self.align_coords[1][0]) - 1 - int(self.align_coords[0][1])

            for txt in txts:
                for e in range(len(txt.exons)-1):
                    intron_span = [int(txt.exons[e][1])+1, int(txt.exons[e+1][0])-1]

                    if novel_intron_span[0] == intron_span[0] and novel_intron_span[1] == intron_span[1]:
                        self.novel = False
                        break

                if not self.novel:
                    break

        elif 'AS' in self.type and self.edge == 'left':
            for txt in txts:
                for exon in txt.exons:
                    if int(self.align_coords[0]) == int(exon[0]):
                        self.novel = False
                        break

                if not self.novel:
                    break

        elif 'AS' in self.type and self.edge == 'right':
            for txt in txts:
                for exon in txt.exons:
                    if int(self.align_coords[1]) == int(exon[1]):
                        self.novel = False
                        break

                if not self.novel:
                    break

        return novel_events

    def longest_orf(self):
        orf = translate(self.align.contig.sequence, full=True)
        if orf:
            coverage = "%.2f" % ((float(orf[1]) - float(orf[0]) + 1) / float(len(self.align.contig.sequence)))
            self.orf = "%d-%d;%s;%s..%s;%s" % (orf[0], orf[1], coverage, orf[-1][:3], orf[-1][-3:], orf[2])
        else:
            self.orf = 'na'

        return None

    def find_pep_change(self, refseq):
        if self.type == 'novel_transcript':
            return self.longest_orf()
        
        coord = re.split('[:-]', self.coordinate())

        variant = None

        if self.type == 'novel_utr':
            #include the block that's 'matching' (partially)
            if min(self.align_blocks) == 1:
                b1 = self.align_blocks[0]
                b2 = self.align_blocks[-1] + 1
            else:
                b1 = self.align_blocks[0] - 1
                b2 = self.align_blocks[-1]
            qcoord1 = self.align.query_blocks[b1-1][0]
            qcoord2 = self.align.query_blocks[b2-1][1]
        else:
            qcoord1 = self.align.query_blocks[self.align_blocks[0]-1][0]
            qcoord2 = self.align.query_blocks[self.align_blocks[-1]-1][1]
        if qcoord1 < qcoord2:
            variant = self.align.contig.sequence[qcoord1-1:qcoord2]
        else:
            variant = self.align.contig.sequence[qcoord2-1:qcoord1]
            variant = reverse_complement(variant)
                         
        cdna_original = self.construct_cdna(coord, self.txt, refseq)
        cdna_changed = self.construct_cdna(coord, self.txt, refseq, variant=variant, change=self.type, exons=self.exons)

        #print 'original:', cdna_original
        #print 'changed:', cdna_changed

        frame = 0
        if self.type == 'novel_utr':
            frame = None

        if self.txt.strand == '+':
            pep_original = translate(cdna_original, orient='+', frame=0)
            pep_changed = translate(cdna_changed, orient='+', frame=frame)
        else:
            pep_original = translate(cdna_original, orient='-', frame=0)
            pep_changed = translate(cdna_changed, orient='-', frame=frame)

        if not pep_changed or not pep_original:
            return 'na'
        
        #print 'original', pep_original
        #print 'changed', pep_changed

        mutation = pep_change(pep_original, pep_changed)

        self.orf = mutation

    def construct_cdna(self, coord, txt, refseq, variant=None, change=None, exons=None):
        cdna = ""
                    
        for i in range(len(txt.exons)):
            if txt.coding_type() != 'CODING' or not overlap(txt.exons[i], [txt.cdsStart, txt.cdsEnd]):
                continue
                
            if subsume([txt.cdsStart, txt.cdsStart], txt.exons[i]):
                start = int(txt.cdsStart) + 1
            else:
                start = txt.exons[i][0]

            if subsume([txt.cdsEnd, txt.cdsEnd], txt.exons[i]):
                end = int(txt.cdsEnd)
            else:
                end = int(txt.exons[i][1])

            exon = refseq.GetSequence(coord[0], int(start), int(end))
            
            if change:
                if change.lower() == 'retained_intron' and i+1 == exons[0]:
                    intron = refseq.GetSequence(coord[0], end+1, txt.exons[i+1][0]-1)
                    exon += intron

                elif change.lower() == 'novel_exon' and i+1 == exons[0]:
                    exon += variant

                elif change.lower() == 'skipped_exon' and i+1 in exons:
                    exon = ''

                elif change.lower() == 'novel_intron' and i+1 == exons[0]:
                    bases_deleted = int(coord[1])-int(start), int(coord[2])-int(start)
                    new_exon = exon[:bases_deleted[0]] + exon[bases_deleted[1]+1:]
                    exon = new_exon

                elif change.lower() in ['as5', 'as3', 'as53'] and i+1 == exons[0]:
                    new_exon = refseq.GetSequence(coord[0], int(coord[1]), int(coord[2]))
                    exon = new_exon

                elif change.lower() == 'novel_utr' and i+1 == exons[0]:
                    exon = variant
            
            cdna += exon

        return cdna
