from optparse import OptionParser
import sys, os, re, glob
from utilities.align_parsers import blat
from utilities import intspan, tools
from utilities import track
from utilities.alignment import Alignment
from utilities.bam import BAM
from feature import FeatureFinder
from transcript import Transcript
from cnv import CNV
from copy import deepcopy

VERSION = '1.3.2'

import utilities.cfg
PACKAGE_DIR = "/".join(os.path.abspath(__file__).split("/")[:-2])
PYTHON = 'python'

class Fusion:

    def __init__(self, a1, a2):
        self.align1 = a1
        self.align2 = a2

	if not re.match('^(chr|scaffold)', self.align1.target, re.IGNORECASE):
	    self.align1.target = 'chr' + self.align1.target
	if not re.match('^(chr|scaffold)', self.align2.target, re.IGNORECASE):
            self.align2.target = 'chr' + self.align2.target
	    
        self.contig = a1.query
        self.contig_size = a1.query_len

        self.gene1 = self.gene2 = "na"
	self.orient1 = self.orient2 = "na"
        self.txt1 = self.txt2 = None

        self.overlap_query = None
        self.overlap_query_fraction = None
        self.overlap_target = None
        self.overlap_target_fraction = None
        self.contig_cover = None
        self.contig_cover_fraction = None
        self.num_read_pairs = 0
        self.num_spanning_reads = 0
	
        self.mutation = 'na'
	self.breakpoint = 'na'
	self.lsr_region = None
	self.cnv = None
	
        #filtering
        self.exclude = False
	self.good_read_pair = False
	self.good_span_read = False
        self.read_support = False
        self.gene_fusion = False
        self.lsr = False
        self.contig_local = None
        self.local = None
        self.ff = None
	
	
    def details(self, tsv=True, id=None):
        ctg_fields = self.contig.split(" ")
        ctg = "%s" % (ctg_fields[0])
        target = "%s:%s-%s,%s:%s-%s" % (self.align1.target, self.align1.tstart, self.align1.tend, self.align2.target, self.align2.tstart, self.align2.tend)
        query = "%s-%s,%s-%s" % (self.align1.qstart, self.align1.qend, self.align2.qstart, self.align2.qend)
	
	#defines gene orientation
        orient = "NA,NA"
        if self.align1.strand:
            orient = "%s,%s" % (self.align1.strand, self.align2.strand)
	    
	#defined alignment params
	align_len1 = float(self.align1.qend) - float(self.align1.qstart) + 1
        align_len2 = float(self.align2.qend) - float(self.align2.qstart) + 1
        align_frac1 = align_len1/float(self.align1.query_len)
        align_frac2 = align_len2/float(self.align2.query_len)
        params = "TO:%.2f,CO:%.2f,CC:%.2f,I1:%.1f,I2:%.1f,AF1:%.2f,AF2:%.2f" % (self.overlap_target_fraction, self.overlap_query_fraction, self.contig_cover_fraction, float(self.align1.identity), float(self.align2.identity), align_frac1, align_frac2)
	    
	#defines fusion type
	fusion_type = 'NA'
	if self.gene_fusion:
	    fusion_type = 'gene_fusion'
	elif self.lsr:
	    fusion_type = 'lsr'
	elif self.local:
	    fusion_type = 'local'


        data = [str(ctg), str(self.contig_size), target, query, orient, str(self.num_read_pairs), str(self.num_spanning_reads), self.mutation, self.breakpoint, params]
	
        if id:
            data.insert(0, id)
	    
        if self.gene1:
            data.insert(-1, "%s(%s)%s(%s)" % (self.gene1, self.orient1, self.gene2, self.orient2))
            		
	if self.cnv:
	    data.append(self.cnv)
	    
	data.append(fusion_type)
            
        if tsv:
            return "\t".join(data)
	
    def report_align_params(self):
	align_len1 = float(self.align1.qend) - float(self.align1.qstart) + 1
        align_len2 = float(self.align2.qend) - float(self.align2.qstart) + 1
        align_frac1 = align_len1/float(self.align1.query_len)
        align_frac2 = align_len2/float(self.align2.query_len)

        params = "TO:%.2f,CO:%.2f,CC:%.2f,I1:%.1f,I2:%.1f,AF1:%.2f,AF2:%.2f" % (self.overlap_target_fraction, self.overlap_query_fraction, self.contig_cover_fraction, float(self.align1.identity), float(self.align2.identity), align_frac1, align_frac2)
	
	return params

    def find_read_pairs(self, bam, coord_buffer=0, keep_mito=False):
        if not keep_mito and (self.align1.target == "chrM" or self.align2.target == "chrM" or self.align1.target == "chrMT" or self.align2.target == "chrMT"):
            return

	#determining breakpoint and passing it to determine valid read pairs
	breakpoint = None
	if self.breakpoint.lower() != 'na':
	    chr1, coord1, chr2, coord2 = re.split('[:\|]', self.breakpoint)
	    	    
	    #match up breakpoint with alignment
	    if chr1 != chr2:
		if chr1 == self.align1.target:
		    breakpoint = [chr1, int(coord1)], [chr2, int(coord2)]
		else:
		    breakpoint = [chr2, int(coord2)], [chr1, int(coord1)]
	    else:
		if (int(coord1) == int(self.align1.tstart) or int(coord1) == int(self.align1.tend)) and int(coord1) != int(self.align2.tstart) and int(coord1) != int(self.align2.tend):
		    breakpoint = [chr1, int(coord1)], [chr2, int(coord2)]
		elif (int(coord2) == int(self.align1.tstart) or int(coord2) == int(self.align1.tend)) and int(coord2) != int(self.align2.tstart) and int(coord2) != int(self.align2.tend):
		    breakpoint = [chr2, int(coord2)], [chr1, int(coord1)]
		    
	
	    #add buffer to breakpoint
	    if breakpoint:
		if int(self.align1.tstart) == int(breakpoint[0][1]):
		    region1 = [self.align1.target, int(self.align1.tstart), max(int(self.align1.tend), breakpoint[0][1] + coord_buffer)]
		elif int(self.align1.tend) == breakpoint[0][1]:
		    region1 = [self.align1.target, min(int(self.align1.tstart), breakpoint[0][1] - coord_buffer), int(self.align1.tend)]
	
		if int(self.align2.tstart) == int(breakpoint[1][1]):
		    region2 = [self.align2.target, int(self.align2.tstart), max(int(self.align2.tend), breakpoint[1][1] + coord_buffer)]
		elif int(self.align2.tend) == breakpoint[1][1]:
		    region2 = [self.align2.target, min(int(self.align2.tstart), breakpoint[1][1] - coord_buffer), int(self.align2.tend)]
		
	    else:
		region1 = [self.align1.target, int(self.align1.tstart), int(self.align1.tend)]
		region2 = [self.align2.target, int(self.align2.tstart), int(self.align2.tend)]
	    
	else:
	    region1 = [self.align1.target, int(self.align1.tstart), int(self.align1.tend)]
	    region2 = [self.align2.target, int(self.align2.tstart), int(self.align2.tend)]
	    
	#print self.align1.query, self.align1.tstart, self.align1.tend, self.align2.tstart, self.align2.tend, self.breakpoint, breakpoint, region1, region2
	#make sure after buffer subtraction it's not negative
	if region1[1] < 0:
	    region1[1] = 0
	if region2[1] < 0:
	    region2[1] = 0
		    
	num_read_pairs = bam.confirm(None, None, feature="breakpoint", regions=[region1, region2], breakpoint=breakpoint)	
        self.num_read_pairs = num_read_pairs

    def find_spanning_reads(self, bam, lib, coord_buffer=4):
        if int(self.align1.qstart) > int(self.align2.qstart):
            start = int(self.align2.qend)
            end = int(self.align1.qstart)
        else:
            start = int(self.align1.qend)
            end = int(self.align2.qstart)

        if start > end:
            tmp = start
            start = end
            end = tmp

        start -= coord_buffer
        end += coord_buffer

        if start <= 0:
            start = 1
        if end > int(self.align1.query_len):
            end = int(self.align1.query_len)

	contig = tools.bamfile_contig_name(self.align1.query, lib)
            
        self.num_spanning_reads = bam.confirm(target=contig, pos=[start,end], feature="breakpoint")
	
    def is_local(self, model=None, annot_file=None, local_gap=5000):
        if self.align1.target != self.align2.target:
            self.local = False
            return self.local
        
        if self.align1.target == self.align2.target:
	    spans = [[int(self.align1.tstart), int(self.align1.tend)], [int(self.align2.tstart), int(self.align2.tend)]]
	    if int(self.align1.tstart) > int(self.align2.tstart):
		spans.reverse()
	    spans[0][1] += local_gap

	    if intspan.overlap(spans[0], spans[1]):
		self.local = True
		self.exclude = True

        if self.gene1 and self.gene2 and self.gene1.lower() != 'na' and self.gene2.lower() != 'na':
            if not self.local:
                if self.gene1 == self.gene2:
                    self.local = True
                    self.exclude = True

            if not self.local and model and annot_file:
                if not self.txt1:
                    txts = Transcript.find_by_name(model, annot_file, self.gene1)
                    if txts:
                        txts.sort(lambda x,y:y.length - x.length)
                        self.txt1 = txts[0]

                if not self.txt2:
                    txts = Transcript.find_by_name(model, annot_file, self.gene2)
                    if txts:
                        txts.sort(lambda x,y:y.length - x.length)
                        self.txt2 = txts[0]

                if self.txt1 and self.txt2 and self.txt1.chrom == self.txt2.chrom and intspan.overlap((self.txt1.txStart, self.txt1.txEnd), (self.txt2.txStart, self.txt2.txEnd)):
                    self.local = True
                    self.exclude = True

        return self.local

    def define_mutation(self):
	if self.align1.target != self.align2.target:
	    self.mutation = 'translocation'
	    
	    span1 = [int(self.align1.tstart), int(self.align1.tend)]
	    span2 = [int(self.align2.tstart), int(self.align2.tend)]
	    if self.align1.strand == '-':
		span1.reverse()
	    if self.align2.strand == '-':
		span2.reverse()
			
	    if self.align1.target < self.align2.target:
		if int(self.align1.qstart) < int(self.align2.qstart):
		    self.breakpoint = "%s:%s|%s:%s" % (self.align1.target, span1[1], self.align2.target, span2[0])
		else:
		    self.breakpoint = "%s:%s|%s:%s" % (self.align1.target, span1[0], self.align2.target, span2[1])
		    
		span1.sort(lambda x,y: x-y)
		span1.sort(lambda x,y: x-y)
		self.lsr_region = [[self.align1.target, span1[0], span1[1]], [self.align2.target, span2[0], span2[1]]]
		 
	    else:
		if int(self.align1.qstart) < int(self.align2.qstart):
		    self.breakpoint = "%s:%s|%s:%s" % (self.align2.target, span2[0], self.align1.target, span1[1])
		else:
		    self.breakpoint = "%s:%s|%s:%s" % (self.align2.target, span2[1], self.align1.target, span1[0])
		    
		span1.sort(lambda x,y: x-y)
		span1.sort(lambda x,y: x-y)
		self.lsr_region = [[self.align2.target, span2[0], span2[1]], [self.align1.target, span1[0], span1[1]]]
		
	else:
	    if int(self.align1.qstart) < int(self.align2.qstart):
		span1 = [int(self.align1.tstart), int(self.align1.tend)]
		span2 = [int(self.align2.tstart), int(self.align2.tend)]
		if self.align1.strand == '-':
		    span1.reverse()
		if self.align2.strand == '-':
		    span2.reverse()
	    else:
		span1 = [int(self.align2.tstart), int(self.align2.tend)]
		span2 = [int(self.align1.tstart), int(self.align1.tend)]
		if self.align2.strand == '-':
		    span1.reverse()
		if self.align1.strand == '-':
		    span2.reverse()
	    
	    ascend1 = None
	    if span1[0] < span1[1]:
		ascend1 = True
	    else:
		ascend1 = False
		    
	    ascend2 = None
	    if span2[0] < span2[1]:
		ascend2 = True
	    else:
		ascend2 = False
	
	    if span1[0] < span2[0]:
		if ascend1 and ascend2:
		    self.mutation = 'deletion'
		elif not ascend1 and not ascend2:
		    self.mutation = 'duplication'
		elif ascend1 and not ascend2:
		    self.mutation = 'inversion'
		elif not ascend1 and ascend2:
		    self.mutation = 'inversion'
		    
		self.breakpoint = "%s:%s|%s:%s" % (self.align1.target, span1[1], self.align2.target, span2[0])
		self.lsr_region = [self.align1.target, span1[1], span2[0]]
		
	    elif span1[0] > span2[0]:
		if ascend1 and ascend2:
		    self.mutation = 'duplication'
		elif not ascend1 and not ascend2:
		    self.mutation = 'deletion'
		elif ascend1 and not ascend2:
		    self.mutation = 'inversion'
		elif not ascend1 and ascend2:
		    self.mutation = 'inversion'
		
		self.breakpoint = "%s:%s|%s:%s" % (self.align2.target, span2[0], self.align1.target, span1[1])
		self.lsr_region = [self.align1.target, span2[0], span1[1]]
		
	    #probably not real, mis-assembly
	    elif span1[0] == span2[0]:
		self.mutation = 'NA'
		self.breakpoint = 'NA'
		self.lsr_region = []

		
class FusionFinder:

    def __init__(self, params, debug=False, unique=False, is_multi_mapped=True):
        self.params = params
        self.fusions = []
        self.debug = debug
        self.unique = unique

        self.filtering = {'min_id':98, 'overlap':True, 'chrM':True, 'haplotype':True, 'readpairs':[0,0], 'min_span_reads':0, 'multimap':False, 'gene_fusions':False}
	
	if params and params.has_key('min_identity'):
	    self.filtering['min_id'] = params['min_identity']
	#FeatureFinder
	self.ff = None

        #debug option
        if debug:
            print self.params
	    
	self.is_multi_mapped = is_multi_mapped
	
	self.fusion_groups = {}

    def get_split_aligns(self, aligns):
        "identifies alignment pairs exhibiting characteristics expected in a true fusion"
	
	fusions = []

        if self.debug:
            print ''
            for i in range(len(aligns)):
                print i,aligns[i].score,aligns[i].identity,aligns[i].match_len,aligns[i].query,aligns[i].qstart,aligns[i].qend,aligns[i].target,aligns[i].tstart,aligns[i].tend
            print "---"
        
        skip = None
        #for keeping track of which of the best alignments used
        count = 0

        #iterate over the alignment pair combos (top 5 alignments currently kept)
        for i in range(len(aligns)):
            #keep track of how many alignments to use
            if i == 0 or aligns[i].query != aligns[i-1].query:
                count = 1
            else:
                count += 1
                if count > self.params['num_best']:
                    continue

            
            if skip and aligns[i].query == aligns[i-1].query:
                continue
            else:
                skip = False

            #check identity
            if aligns[i].identity < self.params['min_identity']:
                if self.debug:
                    print "failed identity", i, aligns[i].identity,  self.params['min_identity']
                continue 

            #check if any member of the alignment set has a single alignment over 95% of the total length; if so, skip
            if (float(aligns[i].qend) - float(aligns[i].qstart) + 1.0) / float(aligns[i].query_len) >= self.params['max_match_fraction']:
                if self.debug:
                    print "skip (major alignment)",aligns[i].query
                skip = True
                continue
            

            if skip:
                continue
             
            for j in range(i+1,len(aligns)):
                if not aligns[j].query == aligns[i].query:
                    break
		
		#check identity
                if aligns[j].identity < self.params['min_identity']:
                    if self.debug:
                        print "failed identity", j, aligns[j].identity,  self.params['min_identity']
                    continue

		
		#check if any member of the alignment set has a single alignment over 95% of the total length; if so, skip
                if (float(aligns[j].qend) - float(aligns[j].qstart)) / float(aligns[j].query_len) >= self.params['max_match_fraction']:
                    if self.debug:
                        print "second alignment too big", i, j, aligns[i].query
                    break

                
                #if one of the alignments is a subset of the other
                if intspan.subsume([aligns[i].qstart, aligns[i].qend], [aligns[j].qstart, aligns[j].qend]) or intspan.subsume([aligns[j].qstart, aligns[j].qend], [aligns[i].qstart, aligns[i].qend]):
                    continue

                #Check how much the query alignments between the contigs overlap
                (overlap_query, overlap_query_fraction) = self.overlap([aligns[i].qstart, aligns[i].qend], [aligns[j].qstart, aligns[j].qend], aligns[i].query_len)
		(overlap_q1, overlap_q1_fraction) = self.overlap([aligns[i].qstart, aligns[i].qend], [aligns[j].qstart, aligns[j].qend], int(aligns[i].qend) - int(aligns[i].qstart) + 1)
		(overlap_q2, overlap_q2_fraction) = self.overlap([aligns[i].qstart, aligns[i].qend], [aligns[j].qstart, aligns[j].qend], int(aligns[j].qend) - int(aligns[j].qstart) + 1)

                #Check if the overlap query between the contigs is too high
		if not overlap_query or overlap_query_fraction < self.params['max_query_olap'] and overlap_q1_fraction < self.params['max_query_olap_single'] and overlap_q2_fraction < self.params['max_query_olap_single']:
		#if not overlap_query or overlap_query_fraction < self.params['max_query_olap'] and overlap_q1_fraction < 0.3 and overlap_q2_fraction  < 0.3:
                    #Get the target query overlap
                    (overlap_target, overlap_target_fraction) = self.overlap([aligns[i].tstart, aligns[i].tend], [aligns[j].tstart, aligns[j].tend], aligns[i].query_len)
                    
                    #if targets are not the same, means no overlap
                    if aligns[i].target == aligns[j].target:
                        (overlap_target, overlap_target_fraction) = self.overlap([aligns[i].tstart, aligns[i].tend], [aligns[j].tstart, aligns[j].tend], aligns[i].query_len)
                    else:
                        (overlap_target, overlap_target_fraction) = (0.0, 0.0)

                    #Check if the target overlap between the contigs is too high
                    if overlap_target_fraction > self.params['max_target_olap']:
                        if not aligns[i].strand:
                            continue
                        elif aligns[i].strand == aligns[j].strand:
                            continue

                    #Get the coverage of both alignments
                    (contig_cover, contig_cover_fraction) = self.combined_aligned(aligns[i], aligns[j], overlap_query)
                    #(contig_cover, contig_cover_fraction) = self.query_spanned(aligns[i], aligns[j])

                    #Check if the cover enough of the contig between the two alignments is too low
                    if contig_cover_fraction >= self.params['min_fusion_fraction']:
                        if self.debug:
                            print "good",i,j,aligns[i].query_len,aligns[i].match_len,float(aligns[i].match_len) / float(aligns[i].query_len),aligns[i].query,aligns[i].score,aligns[i].identity,aligns[i].target,aligns[i].tstart,aligns[i].tend,aligns[j].query,aligns[j].score,aligns[j].identity,aligns[j].target,aligns[j].tstart,aligns[j].tend,overlap_target,overlap_target_fraction,aligns[i].qstart,aligns[i].qend,aligns[j].qstart,aligns[j].qend,overlap_query,overlap_query_fraction,contig_cover_fraction

                        #Create a fusion object as we've passed all the filters
                        fusion = Fusion(aligns[i], aligns[j])
                        fusion.overlap_query_fraction = overlap_query_fraction
                        fusion.overlap_target_fraction = overlap_target_fraction
                        fusion.contig_cover_fraction = contig_cover_fraction
                        fusions.append(fusion)
                        
                    else:
                        if self.debug:
                            print "failed contig_coverage:",i,j,aligns[i].query_len,aligns[i].match_len,float(aligns[i].match_len) / float(aligns[i].query_len),aligns[i].query,aligns[i].score,aligns[i].identity,aligns[i].target,aligns[i].tstart,aligns[i].tend,aligns[j].query,aligns[j].score,aligns[j].identity,aligns[j].target,aligns[j].tstart,aligns[j].tend,overlap_target,overlap_target_fraction,aligns[i].qstart,aligns[i].qend,aligns[j].qstart,aligns[j].qend,overlap_query,overlap_query_fraction,contig_cover_fraction
                elif self.debug:
                    print "failed query_olap:",i,j,aligns[i].query_len,aligns[i].match_len,float(aligns[i].match_len) / float(aligns[i].query_len),aligns[i].query,aligns[i].score,aligns[i].identity,aligns[i].target,aligns[i].tstart,aligns[i].tend,aligns[j].query,aligns[j].score,aligns[j].identity,aligns[j].target,aligns[j].tstart,aligns[j].tend,aligns[i].qstart,aligns[i].qend,aligns[j].qstart,aligns[j].qend,overlap_query,overlap_query_fraction

	return fusions

    def overlap(self, span1, span2, total_len):
        "calculates overlap between two ranges and gives a proportion of the total length covered"
        
        s1 = [int(span1[0]), int(span1[1])]
        s2 = [int(span2[0]), int(span2[1])]
        s1.sort(lambda x,y:x-y)
        s2.sort(lambda x,y:x-y)

        olap = intspan.overlap(s1, s2)
        olap_fraction = 0.0
        if olap:
            olap_fraction = float(olap) / float(total_len)

        return (olap, olap_fraction)


    def combined_aligned(self, align1, align2, olap):
        "calculate combined fractions of query aligned"
        
        align1_aligned =  int(align1.qend) - int(align1.qstart) + 1
        align2_aligned =  int(align2.qend) - int(align2.qstart) + 1
        total_aligned = align1_aligned + align2_aligned - olap
        aligned_fraction = float(total_aligned) / float(align1.query_len)

        return (total_aligned, aligned_fraction)

    def query_spanned(self, align1, align2):
        coords = [int(align1.qstart), int(align1.qend), int(align2.qstart), int(align2.qend)]
        span = max(coords) - min(coords) + 1
        fraction = float(span) / float(align1.query_len)

        return (span, fraction)
    
    def find(self, file):
        "extract fusion partners"
        
        #filter alignments
        sys.stderr.write("parsing alignments...")
        aligns = []

        #filters = {'count':self.params['num_aligns'], 'identity':self.params['min_identity']}
        filters = {'bestn':self.params['num_aligns']}	
	        
        ext = os.path.splitext(file)[1]
        if ext == ".psl":
            aligns = blat.parse(file, filters, noblocks=True, noline=True)
        elif ext == ".gff":
            aligns = exonerate.parse(file, filters)
        else:
            return False
        
        #set orientation if information is there
        #set strand of alignments
        for a in aligns:
            if a.target_strand:
                a.strand = a.target_strand
            elif a.query_strand:
                a.strand = a.query_strand
            else:
                a.strand = None

        sys.stderr.write("done\n")

        #identify splitpairs
        sys.stderr.write("extracting split alignments...")
        fusions = self.get_split_aligns(aligns)
        sys.stderr.write("done\n")

        #only keep unique pairs and different partners
        if self.unique:
            fusion_by_contig = {}
            same_partners = {}
            for fusion in fusions:
                if fusion_by_contig.has_key(fusion.contig):
                    fusion_by_contig[fusion.contig] += 1
                else:
                    fusion_by_contig[fusion.contig] = 1

                if fusion.align1.target == fusion.align2.target:
                    same_partners[fusion.contig] = True

            unique_fusions = [fusion for fusion in fusions if fusion_by_contig[fusion.contig] == 1 and not same_partners.has_key(fusion.contig)]
            self.fusions.extend(unique_fusions)
	    
	else:
	    self.fusions.extend(fusions)

    def confirm_reads(self, genome_bam=None, library=None, contigs_bam=None, coord_buffer=0, keep_mito=False):
        if genome_bam:
            sys.stderr.write("finding read pair evidence...")
            for fusion in self.fusions:
                fusion.find_read_pairs(genome_bam, coord_buffer=coord_buffer, keep_mito=keep_mito)

            sys.stderr.write("done\n")

	if contigs_bam:
            sys.stderr.write("finding reads across breakpoint...")
            for fusion in self.fusions:
                fusion.find_spanning_reads(contigs_bam, library)

            sys.stderr.write("done\n")
	    
    @classmethod
    def compare_lsr(cls, fus1, fus2):
	#sort by read_score, then by coordinate
	#print fus1.align1.query, fus1.num_read_pairs, fus1.num_spanning_reads 
	if fus1.num_read_pairs >= 0 and fus1.num_spanning_reads >= 0:
	    fus1.read_score = 0.2*fus1.num_read_pairs + 5*fus1.num_spanning_reads
	if fus2.num_read_pairs >= 0 and fus2.num_spanning_reads >= 0:
	    fus2.read_score = 0.2*fus2.num_read_pairs + 5*fus2.num_spanning_reads
	
	if fus1.read_score and fus2.read_score:
	    if fus1.read_score > fus2.read_score:
		return -1
	    elif fus1.read_score < fus2.read_score:
		return 1
	    else:
		return FusionFinder.compare_coord(fus1, fus2)
	else:
	    return FusionFinder.compare_coord(fus1, fus2)
    
    @classmethod
    def compare_coord(cls, fus1, fus2):
	if fus1.align1.target < fus1.align2.target:
	    align1 = [fus1.align1, fus1.align2]
	elif fus1.align1.target > fus1.align2.target:
	    align1 = [fus1.align2, fus1.align1]
	elif int(fus1.align1.tstart) < int(fus1.align2.tstart):
	    align1 = [fus1.align1, fus1.align2]
	elif int(fus1.align2.tstart) < int(fus1.align1.tstart):
	    align1 = [fus1.align2, fus1.align1]
	else:
	    align1 = [fus1.align1, fus1.align2]
	    
	if fus2.align1.target < fus2.align2.target:
	    align2 = [fus2.align1, fus2.align2]
	elif fus2.align1.target > fus2.align2.target:
	    align2 = [fus2.align2, fus2.align1]
	elif int(fus2.align1.tstart) < int(fus2.align2.tstart):
	    align2 = [fus2.align1, fus2.align2]
	elif int(fus2.align2.tstart) < int(fus2.align1.tstart):
	    align2 = [fus2.align2, fus2.align1]
	else:
	    align2 = [fus2.align1, fus2.align2]
	    
	if align1[0].target < align2[0].target:
	    return -1
	elif align1[0].target > align2[0].target:
	    return 1
	elif align1[1].target < align2[1].target:
	    return -1
	elif align1[1].target > align2[1].target:
	    return 1
	elif int(align1[0].tstart) < int(align2[0].tstart):
	    return -1
	elif int(align1[0].tstart) > int(align2[0].tstart):
	    return 1
	elif int(align1[1].tstart) < int(align2[1].tstart):
	    return -1
	elif int(align1[1].tstart) > int(align2[1].tstart):
	    return 1
	elif int(align1[0].tend) < int(align2[0].tend):
	    return -1
	elif int(align1[0].tend) > int(align2[0].tend):
	    return 1
	elif int(align1[1].tend) < int(align2[1].tend):
	    return -1
	elif int(align1[1].tend) > int(align2[1].tend):
	    return 1
	else:
	    return 0
	
    def report(self, outdir, append=False, filtering=False, cnv_file=None):
        if append:
            mode = 'a'
        else:
            mode = 'w'

        unfiltered_outfile = outdir + "/fusions.tsv"
        filtered_outfile = outdir + "/fusions_filtered.tsv"
        local_outfile = outdir + "/local.tsv"

        column_header = ['id', 'contig', 'contig_size', 'genomic_regions', 'contig_regions', 'strands', 'read_pairs', 'spanning_reads', 'rearrangement', 'breakpoint', 'genes', 'alignment_params', 'type']
	
	if cnv_file:
	    column_header.insert(11, "cnv")
	    
	if not filtering:
            if mode == 'w' and os.path.exists(unfiltered_outfile):
                os.remove(unfiltered_outfile)
            out = open(unfiltered_outfile, mode)

            fusions = self.fusions[:]
            fusions.sort(lambda x,y: int(y.align1.query_len) - int(x.align1.query_len))
	    
	    out.write('\t'.join(column_header[1:]) + '\n')
            self.output(out, fusions)
	    	    	
	else:
	    fusion_groups = deepcopy(self.fusion_groups)
	    remove_bp = []
	    for bp, fusions in fusion_groups.iteritems():
		remove = []
		for i in range(len(fusions)):
		    if fusions[i].exclude or not fusions[i].read_support:
			remove.append(i)
			
		remove.sort(lambda x,y: y-x)
		for i in remove:
		    del fusions[i]
		
		if not fusions:
		    remove_bp.append(bp)
	
	    for bp in remove_bp:
		del fusion_groups[bp]

	    outfile = filtered_outfile
	    if mode == 'w' and os.path.exists(outfile):
		os.remove(outfile)
	    out = open(outfile, mode)

	    first_members = [group[0] for group in fusion_groups.values()]
	    first_members.sort(FusionFinder.compare_lsr)
	    breakpoints_ordered = [fusion.breakpoint for fusion in first_members]
	
	    out.write('\t'.join(column_header) + '\n')
	    self.output_by_breakpoints(fusion_groups, out, order=breakpoints_ordered)

	    #local
	    if mode == 'w' and os.path.exists(local_outfile):
		os.remove(local_outfile)
	    out = open(local_outfile, mode)
	
	    fusion_groups = deepcopy(self.fusion_groups)
	    remove_bp = []
	    for bp, fusions in fusion_groups.iteritems():
		remove = []
		for i in range(len(fusions)):
		    if not (fusions[i].local or fusions[i].contig_local) or not fusions[i].read_support:
			remove.append(i)
			
		remove.sort(lambda x,y: y-x)
		for i in remove:
		    del fusions[i]
		
		if not fusions:
		    remove_bp.append(bp)
	    for bp in remove_bp:
		del fusion_groups[bp]
	
	    first_members = [group[0] for group in fusion_groups.values()]
	    first_members.sort(FusionFinder.compare_lsr)
	    breakpoints_ordered = [fusion.breakpoint for fusion in first_members]
	
	    out.write('\t'.join(column_header) + '\n')
	    self.output_by_breakpoints(fusion_groups, out, order=breakpoints_ordered)
	    
	    	    
    def overlap_cnvs(self, fusions, cnvs):
	for fusion in fusions:
	    if not fusion.lsr_region:
		continue
	    
	    if fusion.mutation != 'translocation':
		fusion.cnv = "%s:%s" % (fusion.lsr_region[0], cnvs.overlap(fusion.lsr_region[0], fusion.lsr_region[1], fusion.lsr_region[2])) 
	    else:
		region1 = fusion.lsr_region[0]
		region2 = fusion.lsr_region[1]
		
		cnv1 = cnvs.overlap(region1[0], region1[1], region1[2])
		cnv2 = cnvs.overlap(region2[0], region2[1], region2[2])
		
		fusion.cnv = "%s:%s;%s:%s" % (region1[0], cnv1, region2[0], cnv2)

    def output(self, out, fusions, tsv=True):
        for fusion in fusions:
            out.write(fusion.details(tsv=tsv) + "\n")
	    
    def add_spanning_reads(self, groups):
	for group, fusions in groups.iteritems():
	    total = 0
	    
	    for fusion in fusions:
		total += fusion.num_spanning_reads
		
	    for fusion in fusions:
		fusion.num_spanning_reads = total
		
    def group_genes(self, fusions):
        gene2pairs = {}
        gene_pairs = {}
        
        for i in range(len(fusions)):
            fusion = fusions[i]

            if fusion.gene1.lower() == 'na' or not fusion.gene1:
                g1 = ' '.join((fusion.align1.target, str(fusion.align1.tstart), str(fusion.align1.tend)))
            else:
                g1 = fusion.gene1

            if fusion.gene2.lower() == 'na' or not fusion.gene2:
                g2 = ' '.join((fusion.align2.target, str(fusion.align2.tstart), str(fusion.align2.tend)))
            else:
                g2 = fusion.gene2

            if g1 < g2:
                gene_pair = g1 + ' ' + g2
            else:
                gene_pair = g2 + ' ' + g1

            if not gene_pairs.has_key(gene_pair):
                gene_pairs[gene_pair] = [fusion]
            else:
                gene_pairs[gene_pair].append(fusion)

        for pair in gene_pairs:
            for gene in pair.split(' '):
                if not gene2pairs.has_key(gene):
                    gene2pairs[gene] = [pair]
                else:
                    gene2pairs[gene].append(pair)
            
        return gene_pairs, gene2pairs
    
    def group_by_breakpoint(self):
	breakpoints = {}
	for fusion in self.fusions:
	    if not breakpoints.has_key(fusion.breakpoint):
		breakpoints[fusion.breakpoint] = [fusion]
	    else:
		breakpoints[fusion.breakpoint].append(fusion)
		
	self.fusion_groups = breakpoints

    def output_by_genes(self, gene_pairs, gene2pairs, out, tsv=True):
        gene_pairs_list = gene_pairs.keys()
        gene_pairs_list.sort(lambda x,y: len(gene_pairs[y]) - len(gene_pairs[x]))
        
        shown = {}
        pair_count = 0
        for pair in gene_pairs_list:
            if not shown.has_key(pair):
                case_count = 1
                pair_count += 1
                for fusion in gene_pairs[pair]:
                    if len(gene_pairs[pair]) == 1:
                        event_id = str(pair_count)
                    else:
                        event_id = "%s.%s" % (pair_count, case_count)
                    out.write(fusion.details(tsv, event_id) + "\n")
                    case_count += 1                  
                shown[pair] = True

                for gene in pair.split(' '):
                    if gene2pairs.has_key(gene):
                        for pair2 in gene2pairs[gene]:
                            
                            if shown.has_key(pair2):
                                continue
                            
                            pair_count += 1
                            case_count = 1
                            for fusion in gene_pairs[pair2]:
                                if len(gene_pairs[pair2]) == 1:
                                    event_id = str(pair_count)
                                else:
                                    event_id = "%s.%s" % (pair_count, case_count)
                                out.write(fusion.details(tsv, event_id) + "\n")
                                case_count += 1
                            shown[pair2] = True
			    
    def output_by_breakpoints(self, breakpoints, out, tsv=True, order=None):
	breakpoints_list = breakpoints.keys()
	breakpoints_list.sort(lambda x,y: len(breakpoints[y]) - len(breakpoints[x]))
	
	if order:
	    breakpoints_list = order
	  
        breakpoint_count = 0
	for breakpoint in breakpoints_list:
	    case_count = 1
	    breakpoint_count += 1
	    for fusion in breakpoints[breakpoint]:
		if len(breakpoints[breakpoint]) == 1:
		    event_id = str(breakpoint_count)
		else:
		    event_id = "%s.%s" % (breakpoint_count, case_count)
		out.write(fusion.details(tsv, event_id) + "\n")
		case_count += 1                  
          
    def parse_fusions_dir(self, path):
	fusion_dirs = sorted(glob.glob(os.path.join(path, '*')))
		
	fusion_files = []
	num_fusion_dirs = 0
	missing_dirs = []
	for job_num in range(1, len(fusion_dirs)+1):
	    cluster_outdir = "%s/%s" % (args[0], job_num)
	    if os.path.isdir(cluster_outdir):
		num_fusion_dirs += 1
		    
		fusion_file = cluster_outdir + '/fusions.tsv'		    
		if os.path.exists(fusion_file):
		    #print fusion_file
		    fusion_files.append(fusion_file)
		else:
		    missing_dirs.append(str(job_num))
		    sys.stdout.write("%s does not have fusion file\n" % (cluster_outdir))
			
	sys.stdout.write("fusion dirs:%s fusion files:%s\n" % (num_fusion_dirs, len(fusion_files)))
		
	if num_fusion_dirs == len(fusion_files):
	    concat_outfile = outdir + "/fusions.tsv"
	    #print concat_outfile
	    concat_out = open(concat_outfile, 'w')
		    
	    #just for check if header is to be written
	    count = 0
	    for fusion_file in fusion_files:
		infile = open(fusion_file, 'r')
		if count == 0:
		    concat_out.writelines(open(fusion_file, 'r').readlines())
		else:
		    concat_out.writelines(open(fusion_file, 'r').readlines()[1:])
		infile.close()
			
		print fusion_file
		self.parse_fusions(fusion_file)
		count += 1
			
	    concat_out.close()
	else:
	    sys.stdout.write("missing (%s):%s\n" % (len(missing_dirs), ','.join(missing_dirs)))
			    
    def parse_fusions(self, fusions_file):
        genes = re.compile(r'(?P<gene1>.+)\((?P<orient1>.+)\)(?P<gene2>.+)\((?P<orient2>.+)\)')
        
	filtered = False
        for line in open(fusions_file, 'r'):
            cols = line.rstrip('\n').split('\t')
	    #empty line
	    if len(cols) < 2:
		continue
	    
	    if filtered:
		cols = cols[1:]

	    if cols[0] == 'contig':
		continue
	    elif cols[0] == 'id':
		filtered = True
		continue

            align1 = Alignment(None)
            align2 = Alignment(None)
	    
	    align1.query = align2.query = cols[0]
	    align1.query_len = align2.query_len = cols[1]

	    targets = cols[2].split(',')
            align1.target, align1.tstart, align1.tend = re.split('[:-]', targets[0])
            align2.target, align2.tstart, align2.tend = re.split('[:-]', targets[1])

            queries = cols[3].split(',')
            align1.qstart, align1.qend = re.split('-', queries[0])
            align2.qstart, align2.qend = re.split('-', queries[1])

            align1.strand, align2.strand = cols[4].split(',')

            fusion = Fusion(align1, align2)
            params = cols[10].split(',')
            fusion.overlap_target_fraction = float(params[0].split(':')[1])
            fusion.overlap_query_fraction = float(params[1].split(':')[1])
            fusion.contig_cover_fraction = float(params[2].split(':')[1])
            align1.identity = float(params[3].split(':')[1])
            align2.identity = float(params[4].split(':')[1])

            fusion.num_read_pairs = int(cols[5])
            fusion.num_spanning_reads = int(cols[6])

            fusion.mutation = cols[7]
	    fusion.breakpoint = cols[8]

            m = genes.search(cols[9])
            if m and m.groupdict():
                fusion.gene1 = m.groupdict()['gene1']
                fusion.orient1 = m.groupdict()['orient1']
                fusion.gene2 = m.groupdict()['gene2']
                fusion.orient2 = m.groupdict()['orient2']
		
		for item in (fusion.gene1, fusion.orient1, fusion.gene1, fusion.orient1):
		    if item.lower() == 'na':
			item = item.upper()
	    
            self.fusions.append(fusion)

    def annotate(self, genome=None, model=None):
        self.ff = FeatureFinder(genome, model)

	for fusion in self.fusions:   
	    if fusion.breakpoint.lower() != 'na':
		target1, bp1, target2, bp2 = re.split('[|:]', fusion.breakpoint)
		gene1, strand1 = self.ff.get_feature(' '.join((target1, bp1, bp1)), gene_only=True, gene_strand=True)
		gene2, strand2 = self.ff.get_feature(' '.join((target2, bp2, bp2)), gene_only=True, gene_strand=True)
	    else:
		gene1 = 'NA'
		gene2 = 'NA'
		
	    fusion.gene1 = gene1
            fusion.gene2 = gene2

            if gene1 and gene1 != 'NA':
                if strand1 == fusion.align1.strand:
                    fusion.orient1 = '+'
                else:
                    fusion.orient1 = '-'
            else:
                fusion.orient1 = 'NA'

            if gene2 and gene2 != 'NA':
                if strand2 == fusion.align2.strand:
                    fusion.orient2 = '+'
                else:
                    fusion.orient2 = '-'
            else:
                fusion.orient2 = 'NA'


    def post_filter(self):
        #sort by contig
        #fusions by contig
        fusions_by_contig = {}
        for i in range(len(self.fusions)):
            if not fusions_by_contig.has_key(self.fusions[i].contig):
                fusions_by_contig[self.fusions[i].contig] = [i]
            else:
                fusions_by_contig[self.fusions[i].contig].append(i)

        for contig, fusion_idx in fusions_by_contig.iteritems():
            contig_local = False
            for idx in fusion_idx:
                fusion = self.fusions[idx]
		
		if not fusion.local:
		    if self.ff and fusion.is_local(self.ff.model, self.ff.annot_file):
			fusion.local = True
			contig_local = True
			break
		    elif fusion.is_local():
			fusion.local = True
			contig_local = True
			break
			                
            if contig_local:
                for idx in fusion_idx:
                    self.fusions[idx].contig_local = True
                    self.fusions[idx].exclude = True

        self.post_filter_others()
	
    def get_bad_gene_pairs(self, infile):
	gene_pairs = {}
	
	for line in open(infile, 'r'):
	    genes = line.rstrip('\n').split()

	    if len(genes) != 2:
		continue
	    
	    if gene_pairs.has_key(genes[0]):
		gene_pairs[genes[0]][genes[1]] = True
	    else:
		gene_pairs[genes[0]] = {genes[1]:True}
		
	    if gene_pairs.has_key(genes[1]):
		gene_pairs[genes[1]][genes[0]] = True
	    else:
		gene_pairs[genes[1]] = {genes[0]:True}

	return gene_pairs
    
    def get_bad_breakpoints(self, infile):
	breakpoints = {}
	
	for line in open(infile, 'r'):
	    breakpoint = line.rstrip('\n')
	    breakpoints[breakpoint] = True

	return breakpoints
    
    def post_filter_others(self):
        exclude_indices = {}
        mappings = {}
        contig2fusions = {}
	
	bad_gene_pairs = None
	if self.filtering.has_key('bad_gene_pairs'):
	    bad_gene_pairs = self.get_bad_gene_pairs(self.filtering['bad_gene_pairs'])
	    
	bad_breakpoints = None
	if self.filtering.has_key('bad_breakpoints'):
	    bad_breakpoints = self.get_bad_breakpoints(self.filtering['bad_breakpoints'])
	    
	for i in range(len(self.fusions)):
            fusion = self.fusions[i]

            if fusion.contig_local or fusion.local:
                fusion.exclude = True
		
	    if not fusion.exclude and self.filtering.has_key('bad_gene_pairs') or self.filtering.has_key('bad_breakpoints'):
		if fusion.gene1.lower() != 'na' and fusion.gene2.lower() != 'na' and bad_gene_pairs.has_key(fusion.gene1) and bad_gene_pairs[fusion.gene1].has_key(fusion.gene2):
		    fusion.exclude = True
		    
		if (fusion.gene1.lower() == 'na' or fusion.gene2.lower() == 'na') and bad_breakpoints.has_key(fusion.breakpoint):
		    fusion.exclude = True
	    
            #basic filtering
            if not fusion.exclude and self.filtering.has_key('chrM') and self.filtering['chrM']:
                if fusion.align1.target.lower() == 'chrm' or fusion.align2.target.lower() == 'chrm' or fusion.align1.target.lower() == 'chrmt' or fusion.align2.target.lower() == 'chrmt':
                    fusion.exclude = True

            if not fusion.exclude and self.filtering.has_key('haplotype') and self.filtering['haplotype']:
                if (re.match('^chr', fusion.align1.target, re.IGNORECASE) and '_' in fusion.align1.target) or (re.match('^chr', fusion.align2.target, re.IGNORECASE) and '_' in fusion.align2.target):
                    fusion.exclude = True
                    

            #multi-mapping
            if self.filtering.has_key('no_multimap') and self.filtering['no_multimap']:
                contig = fusion.align1.query
                if not mappings.has_key(contig):
                    mappings[contig] = {}
                    contig2fusions[contig] = {}
                    
                coord1 = fusion.align1.target + str(fusion.align1.tstart) + str(fusion.align1.tend)
                if not mappings[contig].has_key(coord1):
                    mappings[contig][coord1] = {}
                    contig2fusions[contig][coord1] = [i]
                else:
                    contig2fusions[contig][coord1].append(i)
                    
                if not mappings[contig][coord1].has_key(fusion.align2.target):
                    mappings[contig][coord1][fusion.align2.target] = 1
                else:
                    mappings[contig][coord1][fusion.align2.target] += 1

                coord2 = fusion.align2.target + str(fusion.align2.tstart) + str(fusion.align2.tend)
                if not mappings[contig].has_key(coord2):
                    mappings[contig][coord2] = {}
                    contig2fusions[contig][coord2] = [i]
                else:
                    contig2fusions[contig][coord2].append(i)
                    
                if not mappings[contig][coord2].has_key(fusion.align1.target):
                    mappings[contig][coord2][fusion.align1.target] = 1
                elif coord2 != coord1:
                    mappings[contig][coord2][fusion.align1.target] += 1

            if fusion.num_read_pairs >= self.filtering['readpairs'][0] and fusion.num_read_pairs <= self.filtering['readpairs'][1]:
                fusion.good_read_pair = True
            if fusion.num_spanning_reads >= self.filtering['min_span_reads']:
                fusion.good_span_read = True
            if fusion.good_read_pair and fusion.good_span_read:
                fusion.read_support = True
		
            if self.filtering.has_key('gene_fusions'):
                if fusion.gene1 and fusion.gene2:
                    if fusion.gene1.lower() == 'na' or fusion.gene2.lower() == 'na':
                        fusion.lsr = True
                    elif fusion.gene1 != fusion.gene2:
			if self.filtering.has_key('gene_orient') and self.filtering['gene_orient']:
			    if fusion.orient1 == fusion.orient2:
				fusion.gene_fusion = True
			else:
			    fusion.gene_fusion = True
                    

        #multi_mapping
	if self.filtering['no_multimap']:
	    multi_mapped = {}
	    for contig in mappings.keys():
		for coord in mappings[contig]:
		    if len(mappings[contig][coord]) > 1:
			for target in mappings[contig][coord]:
			    if mappings[contig][coord][target] > 1:
				for i in contig2fusions[contig][coord]:
				    #print self.fusions[i].details()
				    multi_mapped[i] = True
				#print "--"
	    for i in multi_mapped.keys():
		self.fusions[i].exclude = True
        
    def post_process(self, options):
	#define mutation
	for fusion in ff.fusions:
	    fusion.define_mutation()
	
	#group fusions
	self.group_by_breakpoint()
    
	#collect read support
	if options.genome_bam or options.contigs_bam:
	    self.read_support(genome_bamfile=options.genome_bam, contigs_bamfile=options.contigs_bam, min_mapq=options.mapq_genome, sample_type=options.sample_type, library=options.library, keep_mito=options.keep_mito)
	
	#add spanning reads 
	if options.add_spanning_reads:
	    self.add_spanning_reads(ff.fusion_groups)

	#annotation
	if options.gene_model:
	    genome, model = options.gene_model
	    self.annotate(genome=genome, model=model)
	
	if options.cnv_file:
	    cnvs = CNV(options.cnv_file)
	    cnvs.extract()
	    self.overlap_cnvs(self.fusions, cnvs)

    def read_support(self, genome_bamfile=None, contigs_bamfile=None, min_mapq=None, sample_type=None, library=None, keep_mito=False):
	if genome_bamfile and os.path.exists(genome_bamfile):
	    if os.path.isdir(genome_bamfile):
		bamfiles = glob.glob(os.path.join(genome_bam, '*.bam'))
		for bamfile in bamfiles:
		    genome_bam = BAM(bamfile, min_mapq=min_mapq, sample_type=sample_type)
		    self.confirm_reads(genome_bam=genome_bam, coord_buffer=options.coord_buffer, keep_mito=keep_mito)
	    else:
		genome_bam = BAM(genome_bamfile, min_mapq=min_mapq, sample_type=sample_type)
		self.confirm_reads(genome_bam=genome_bam, coord_buffer=options.coord_buffer, keep_mito=keep_mito)
            
	if contigs_bamfile and os.path.exists(contigs_bamfile):
	    contigs_bam = BAM(contigs_bamfile, min_mapq=options.mapq_contigs)
	    self.confirm_reads(contigs_bam=contigs_bam, library=library, coord_buffer=options.coord_buffer)
	
def create_jobs_file(align_dir, outdir, num, params):
    cluster_dir = outdir + "/cluster"
    if not os.path.isdir(cluster_dir):
	os.mkdir(cluster_dir)
	
    jobs_file = cluster_dir + "/jobs"
    sys.stdout.write("creating job file:%s\n" % (num))
    jobs = open(jobs_file, 'w')
    
    script = os.path.abspath(sys.argv[0])
    source = PACKAGE_DIR + "/setup"

    for i in range(1,num+1):
	outdir = cluster_dir + "/" + str(i)
	if not os.path.isdir(outdir):
	    os.mkdir(outdir)
	
	align_file = "%s/seq.%s.psl" % (align_dir, i)
	line = "source %s;%s %s %s %s -C 0 -B %s -b %s -G %s %s" % (source, PYTHON, script, align_file, outdir, options.genome_bam, options.contigs_bam, options.gene_model[0], options.gene_model[1])
	if options.library:
	    line += " -l " + options.library
	    
	for param, val in params.iteritems():
	    line += " --%s %s" % (param, val)
	
	jobs.write(line + '\n')
        
    jobs.close()
    
    return jobs_file


if __name__ == '__main__':
    usage = "Usage: %proj alignment_[file|dir] output_dir"

    parser = OptionParser(usage=usage, version="%prog " + VERSION)
    parser.add_option("-m", "--max_match_fraction", dest="match_fraction", help="maximum match fraction for individual alignment. Default:0.99", default=0.99, type='float')
    parser.add_option("-q", "--max_query_olap", dest="query_olap", help="maximum query overlap. Default:0.05", default=0.05, type='float')
    parser.add_option("-Q", "--max_query_olap_single", dest="query_olap_single", help="maximum single region query overlap. Default:0.3", default=0.3, type='float')
    parser.add_option("-t", "--max_target_olap", dest="target_olap", help="maximum target overlap. Default:0.0", default=0.0, type='float')
    parser.add_option("-f", "--min_fusion_fraction", dest="fusion_fraction", help="minimum fusion fraction. Default:0.90", default=0.90, type='float')
    parser.add_option("-a", "--num_aligns", dest="num_aligns", help="number alignments. Default:20", default=20, type='int')
    parser.add_option("-A", "--num_best", dest="num_best", help="number of best alignments to use. Default:5", default=5, type='int')
    parser.add_option("-i", "--min_identity", dest="min_identity", help="minimum identity. Default:98", type='float', default=98.0)
    parser.add_option("-s", "--min_span_reads", dest="min_span_reads", help="minimum spanning reads. Default:2", type='int', default=2)
    parser.add_option("-p", "--min_read_pairs", dest="min_read_pairs", help="minimum read pairs. Default:4", type='int', default=4)
    parser.add_option("-P", "--max_read_pairs", dest="max_read_pairs", help="maximum read pairs. Default:2000", type='int', default=2000)
    
    parser.add_option("-d", "--debug", dest="debug", help="debug", action="store_true", default=False)
    parser.add_option("-u", "--unique", dest="unique", help="only keep unique pairs (not same target)", action="store_true", default=False)
    parser.add_option("-c", "--coord_buffer", dest="coord_buffer", help="coordinate buffer for extracting reads. Default:2000", default=2000, type='int')
    parser.add_option("-B", "--genome_bam", dest="genome_bam", help="genomic bam file")
    parser.add_option("-b", "--contigs_bam", dest="contigs_bam", help="contigs bam file")
    parser.add_option("-F", "--fusions", dest="fusions", help="fusions input file given", action="store_true", default=False)
    parser.add_option("-l", "--library", dest="library", help="library name (required for using reads-contigs bam file)")
    parser.add_option("-M", "--keep_mito", dest="keep_mito", help="include mitochondria for read pair evidence", action="store_true", default=False)
    parser.add_option("-n", "--no_multimap", dest="no_multimap", help="no multi-mapping of same genomic region allowed", action="store_true", default=False)
    parser.add_option("-C", "--chunk_size", dest="chunksize", help="break up alignments into chunks of size C", type='int', default=1000)
    parser.add_option("-Y", "--mapq_genome", dest="mapq_genome", help="minimum mapping quality for reads to genome. Default:1", type='int', default=1)
    parser.add_option("-y", "--mapq_contigs", dest="mapq_contigs", help="minimum mapping quality for reads to contigs. Default:0", type='int', default=0)
    parser.add_option("-X", "--filter", dest="filter", help="post-filter", action="store_true", default=False)
    parser.add_option("-G", "--gene_model", dest="gene_model", help="gene model used for annotation e.g. hg18 k where k=known_genes, e=ensembl, r=refseq", nargs=2)
    parser.add_option("-O", "--gene_orient", dest="gene_orient", action="store_true", default=False)
    
    parser.add_option("-k", "--cluster", dest="cluster", help="cluster")
    parser.add_option("-v", "--bad_gene_pairs", dest="bad_gene_pairs")
    parser.add_option("-w", "--bad_breakpoints", dest="bad_breakpoints")
    parser.add_option("-V", "--cnv_file", dest="cnv_file")    
    parser.add_option("-S", "--add_spanning_reads", dest="add_spanning_reads", action="store_true", default=False)
    parser.add_option("-T", "--type", dest="sample_type", help="sample type: transcriptome(default), genome, exome - can be ignored", default='transcriptome')
    
    (options, args) = parser.parse_args()

    params = {}
    params['max_match_fraction'] = options.match_fraction
    params['max_query_olap'] = options.query_olap
    params['max_query_olap_single'] = options.query_olap_single
    params['max_target_olap'] = options.target_olap
    params['min_fusion_fraction'] = options.fusion_fraction
    params['min_identity'] = options.min_identity
    params['num_aligns'] = options.num_aligns
    params['num_best'] = options.num_best
        
    tools.output_log(VERSION, sys.argv, [vars(options)], args[1])
    
    if len(args) == 2:	    
        ff = FusionFinder(params, debug=options.debug, unique=options.unique)

        outdir = args[1]
	       
        if options.fusions:
	    if (os.path.isdir(args[0])):
		ff.parse_fusions_dir(args[0])
	    else:
		ff.parse_fusions(args[0])
		                
        else:
            if os.path.isdir(args[0]):                
                align_files = sorted(glob.glob(os.path.join(args[0], '*.*')))
		
		if not options.cluster:
		    for i in range(len(align_files)):
			sys.stdout.write("processing alignment file #%d\n" % (i+1))
			ff.find(align_files[i])
			
		else:
		    jobs_file = create_jobs_file(args[0], args[1], len(align_files), params)
		    print jobs_file			
			

            elif options.chunksize > 0:                
                sys.stderr.write("breaking up alignments into " + str(options.chunksize) + " chunks...")
                num_files = blat.divide(args[0], options.chunksize, outdir)
                sys.stderr.write("done\n")                

                align_name, align_ext = os.path.splitext(os.path.split(args[0])[1])
                outfile = None
                for i in range(1, num_files+1):
                    sys.stderr.write("processing alignment file #%d\n" % (i))
                    align_file = outdir + "/" + align_name + "." + str(i) + align_ext
                    ff.find(align_file)
                
                #clean up intermediate alignment files
                for i in range(1, num_files+1):
                    cmd = "rm " + outdir + "/" + align_name + "." + str(i) + align_ext
                    os.system(cmd)

            else:
                ff.find(args[0])
		
	#defines breakpoint, group, read_support, cnv
	if not options.cluster:
	    ff.post_process(options)
	
	if not options.fusions and not options.cluster:
	    ff.report(outdir, filtering=False)

	#filtering
	if options.filter:
            ff.filtering['no_multimap'] = options.no_multimap
            ff.filtering['readpairs'] = [options.min_read_pairs, options.max_read_pairs]
            ff.filtering['min_span_reads'] = options.min_span_reads
            ff.filtering['gene_fusions'] = True
	    ff.filtering['gene_orient'] = options.gene_orient

	    if options.bad_gene_pairs:
		ff.filtering['bad_gene_pairs'] = options.bad_gene_pairs
	    if options.bad_breakpoints:
		ff.filtering['bad_breakpoints'] = options.bad_breakpoints
		
            ff.post_filter()

	    
	    #output
	    ff.report(outdir, filtering=options.filter, cnv_file=options.cnv_file)

            

