from utilities.track import Track
from utilities.intspan import union
from utilities.bam import BAM
import os, sys
from utilities.tools import bamfile_contig_name
from optparse import OptionParser

def gene_to_txts(gene2txt_file):
    gene2txts = {}
    
    for line in open(gene2txt_file, 'r'):
        gene, txt = line.rstrip('\n').split('\t')[:2]
        #print gene, txt

        if not gene2txts.has_key(gene):
            gene2txts[gene] = [txt]
        else:
            gene2txts[gene].append(txt)

    return gene2txts

def parse_coverage(coverage_file):
    contig2txts = {}
    contig2genes = {}
    
    for line in open(coverage_file, 'r'):
        cols = line.split('\t')
        contigs = cols[7].split(',')

        contig2txts[cols[0]] = contigs

        #assume one contig will be associated to only one transcript
        if cols[1].lower() != 'na':
            if not contig2genes.has_key(cols[1]):
                contig2genes[cols[1]] = {}
        
                for contig in contigs:
                    contig2genes[cols[1]][contig] = True

    return contig2txts, contig2genes

def contig_to_genes(contig2txts, gene2txts=None):
    contig2genes = {}

    if gene2txts:
        for gene in gene2txts.keys():
            contig2genes[gene] = {}
        
            for txt in gene2txts[gene]:
                if contig2txts.has_key(txt):
                    for contig in contig2txts[txt]:
                        contig2genes[gene][contig] = True
    else:
        for gene in contig2txts.keys():
            contig2genes[gene] = {}
            for contig in contig2txts[gene]:
                contig2genes[gene][contig] = True

    return contig2genes

def get_blocks(track_file):
    all_blocks = None
    
    trk = Track(track_file)

    ext = os.path.splitext(track_file)[1]
    if ext == ".gff":
        all_blocks = trk.parse_gff(just_blocks=True)
    elif ext == ".psl":
        all_blocks = trk.parse_psl(minimum=True, just_blocks=True)

    return all_blocks

def find_coverage(contig2genes, blocks, bam, outfile, lib=None):
    '''
    calculates normalized gene coverage = R/L
    where R = sum of lengths of all unique reads mapped to a gene
          L = length of the union of all alignment blocks mapped to a gene
    '''
    
    out = open(outfile, 'w')
    header = ('gene', 'nreads', 'total_read_length', 'union_aligned_block_length', 'normalized_coverage')
    out.write("\t".join(header) + "\n")
    
    for gene in contig2genes.keys():
        if len(contig2genes[gene].keys()) < 1:
            continue

        blocks_gene = []
        reads_gene = {}

        num_contigs = 0
        for contig in contig2genes[gene].keys():
            if blocks.has_key(contig):
                num_contigs += 1
                
                #print gene, contig, blocks[contig]
                blocks_gene.append(blocks[contig])

                bam_contig_name = bamfile_contig_name(contig, lib)

                for read in bam.bam.fetch(bam_contig_name):
                    #print read.qname, read.rlen, read.pos
                    reads_gene[read.qname] = read.rlen

        if len(blocks_gene) < 1:
            continue

        #L
        union_blocks = union(blocks_gene)
        block_length = get_bases(union_blocks)
        
        num_reads = len(reads_gene.keys())

        #R
        reads_length = get_read_total(reads_gene)
        
        cov = '%.2f' % (float(reads_length)/float(block_length))
        #print gene, num_reads, reads_length, block_length, cov
        out.write("\t".join((gene, str(num_reads), str(reads_length), str(block_length), cov)) + "\n")

    out.close()

def get_bases(blocks):
    length = 0
    for i in range(len(blocks)):
        l = blocks[i][1] - blocks[i][0] + 1
        length += l

    return length

def get_read_total(reads):
    length = 0
    
    for read in reads.keys():
        length += reads[read]

    return length


if __name__ == '__main__':
    usage = "Usage: %proj coverage-file bam-file track-file output-file"
    parser = OptionParser(usage=usage)
    parser.add_option("-g", "--gene2transcript", dest="gene2transcript", help="tab-delimited gene to transcript file (gene<tab>transcript)")
    parser.add_option("-l", "--lib", dest="lib", help="library name - only needed if contigs bam file has library name")
    
    (options, args) = parser.parse_args()
    if len(args) == 4:
        contig2txts, contig2genes = parse_coverage(args[0])
        
        gene2txts = None
        if options.gene2transcript:
            gene2txts = gene_to_txts(options.gene2transcript)
            contig2genes = contig_to_genes(contig2txts, gene2txts=gene2txts)

        #get alignment blocks of contigs
        blocks = get_blocks(args[2])

        bam = BAM(args[1])
        outfile = None
        find_coverage(contig2genes, blocks, bam, args[3], lib=options.lib)
    else:
        parser.error("incorrect number of arguments")
