from optparse import OptionParser
import os, sys, re, glob
from snv import SNV
from feature import FeatureFinder
from utilities import tools, intspan
from utilities.assembly import Assembly
from utilities.track import Track
from utilities.bam import BAM
from utilities.align_parsers import blat
import difflib
from utilities.overlap_coord import OverlapCoord
from annotations import repeat, dbsnp
from bubble import BubbleMapper

VERSION = '1.3.2'

import utilities.cfg
PACKAGE_DIR = "/".join(os.path.abspath(sys.argv[0]).split("/")[:-2])
PYTHON = 'python'

class SNVCaller:

    output_headers = [
        'id',
        'type',
        'chr',
        'chr_start',
        'chr_end',
        'ctg',
        'ctg_len',
        'ctg_start',
        'ctg_end',
        'len',
        'ref',
        'alt',
        'event_reads',
        'contig_reads',
        'genome_reads',
        'gene',
        'repeat-length',
        'ctg_strand',
        'from_end',
        'confirm_contig_region',
        'within_simple_repeats',
        'repeatmasker',
        'within_segdup',
        'at_least_1_read_opposite',
        'dbsnp'
    ]

    def __init__(self, align_file=None, contigs_file=None, genome=None, lib='', min_reads=0, debug=False, min_from_end=None, bubble_mapping=None, sample_type='transcriptome', is_track=False):
        self.align_file = align_file
        self.contigs_file = contigs_file
        self.genome = genome
        self.min_reads = min_reads
        self.debug = debug
	self.min_from_end = min_from_end

        if self.genome:
	    self.get_annotation_files()

        self.lib = lib

        self.snvs = []
	self.grouped_snvs = {}
	
	self.bubble_mapping = bubble_mapping
	self.sample_type = sample_type
	
	self.is_track = is_track
	self.feature_finder = None
		
    def get_annotation_files(self):
	package_dir = "/".join(os.path.abspath(sys.argv[0]).split("/")[:-2])
	genome_dir = package_dir + '/annotations/' + self.genome
	
	if not os.path.isdir(genome_dir):
	    sys.exit()
	
	self.splice_motif_file = genome_dir + '/splice_motifs.txt'
	
	if not os.path.exists(self.splice_motif_file):
	    sys.exit()
	    
    def extract(self, cutoff=None, model=None, aligns=None, refseq=None, no_group=False, match_percent=None, identity=None):
        if not refseq:
	    sys.stderr.write("problem extracting refseq - abort\n")
        
        splice_motifs = tools.get_splice_motifs(self.splice_motif_file)
	
        if not aligns:
	    if not self.is_track:
		aligns = blat.parse(self.align_file, {'unique':True, 'bestn':1, 'match':match_percent, 'identity':identity}, splice_motif_file=self.splice_motif_file, refseq=refseq, noline=True)
	    else:
		track = Track(self.align_file)
		aligns = track.parse_psl(splice_motif_file=self.splice_motif_file, refseq=refseq)
			    
	    for align in aligns:
		align.target = tools.proper_chrom(align.target, self.genome) 
		
            #link contig sequence to alignment
            ass = Assembly(self.lib, k=1)
            ass.fasta = self.contigs_file
            contigs = ass.get_contigs(sequence=True)
            contig_dict = dict((contig.num, contig) for contig in contigs)

            for align in aligns:
                if contig_dict.has_key(align.query):
                    align.contig = contig_dict[align.query]

        
        for align in aligns:
	    if self.bubble_mapping and not self.bubble_mapping.is_bubble_mapped_to_contig(align.query):
		print "remove bubble", align.query
		continue
	     
	    if not re.match('^(chr|scaffold)', align.target, re.IGNORECASE):
		align.target = 'chr' + align.target
	    
            snvs = self.get_snvs(align, splice_motifs=splice_motifs, refseq=refseq, cutoff=cutoff)

            for snv in snvs:
                snv.var_len = align.query_len
                snv.from_end = min(int(snv.var_start) - int(align.qstart), int(align.qend) - int(snv.var_end))

                snv.upshift(refseq)
                snv.expand_contig_region(align.contig.sequence, align.query_strand)
                target = snv.ref
				                        
                self.snvs.append(snv)
		
		
	#group snvs
	if not no_group:
	    self.grouped_snvs = self.group(self.snvs)
	    
    def annotate_genes(self, refseq):
	if self.feature_finder:
	    for groups in self.grouped_snvs.values():
		for snvs in groups.values():
		    gene = self.feature_finder.get_feature(' '.join((snvs[0].ref, str(snvs[0].ref_start), str(snvs[0].ref_end))), refseq=refseq, variant=snvs[0].var_seq, change=snvs[0].snv_type)
		
		    for snv in snvs:
			snv.gene = gene
	    
    def overlap_repeats(self):
	repeat_overlaps = repeat.prepare_overlap(self.genome)
	
	for groups in self.grouped_snvs.values():
	    for key, snvs in groups.iteritems():
		print 'processing repeat', key
		overlaps = repeat.find_overlaps({'chrom':snvs[0].ref, 'start':int(snvs[0].ref_start), 'end':int(snvs[0].ref_end)}, repeat_overlaps)

		if overlaps:
		    attrs = {}
		    for repeat_type, types in overlaps.iteritems():
			if repeat_type == 'simple_repeats':
			    attr = 'within_simple_repeats'
			    
			elif repeat_type == 'segdup':
			    attr = 'within_segdup'
			
			elif repeat_type == 'rmsk':
			    attr = 'repeatmasker'
			
			if types:
			    #only report one with shortest name
			    types_sorted = types.keys()
			    types_sorted.sort(lambda x,y: len(x)-len(y))
			    attrs[attr] = types_sorted[0]
			
		    if attrs:
			for snv in snvs:
			    tools.set_attrs(snv, attrs)	  
			    
	for repeat_olap in repeat_overlaps.values():
	    repeat_olap.finish()

    def overlap_dbsnp(self, refseq):	
	for event_type in ('ins', 'del', 'snv'):
	    if not self.grouped_snvs.has_key(event_type):
		continue
	    
	    for key, snvs in self.grouped_snvs[event_type].iteritems():
		known = dbsnp.find_concordants({'type':snvs[0].snv_type, 'chrom':snvs[0].ref, 'start':int(snvs[0].ref_start), 'end':int(snvs[0].ref_end), 'allele':snvs[0].var_seq.lower(), 'size':int(snvs[0].snv_len)}, genome=self.genome, refseq=refseq)
		
		if known:
		    for snv in snvs:
			snv.dbsnp = ','.join(known)
	

    def add_support(self, contigs_bam=None, lib=None, genome_bam=None, outdir=None, from_end=None, specific_type=None, refseq=None):
        if genome_bam:
            for snv in self.snvs:
                snv.genome_read_support(genome_bam, refseq, from_end=from_end)
		
        if contigs_bam:
	    for snv_type, coord_alleles in self.grouped_snvs.iteritems():
		if specific_type and snv_type != specific_type:
		    continue
		
		for coord_allele in coord_alleles:
		    supports = {}
		    
		    #get unique read support for event
		    snvs = self.grouped_snvs[snv_type][coord_allele]
		    for snv in snvs:
			reads = snv.contig_read_support(contigs_bam, lib=lib, get_reads=True, from_end=from_end)
		    
			for read in reads:
			    supports[read.qname] = read

		    #check at least one read is in opposite orientation
		    reverse = 0
		    for read in supports.values():
			if read.is_reverse:
			    reverse += 1
		    
		    at_least_1_read_opposite = False
		    if reverse > 0 and reverse < len(supports.keys()):
			at_least_1_read_opposite = True

		    for snv in self.grouped_snvs[snv_type][coord_allele]:
			snv.nreads_event = len(supports.keys())
			if snv.nreads_event >= self.min_reads:
			    snv.enough_coverage = True
			
			snv.at_least_1_read_opposite = at_least_1_read_opposite
						
			
		    if snv_type != 'snv' and snv.enough_coverage and (self.sample_type != 'genome' or snv.at_least_1_read_opposite) and outdir:
			fa = outdir + "/" + ":".join([snv_type, coord_allele]) + ".fa"
			out = open(fa, 'w')
			for read in supports.values():
			    out.write('>%s\n%s\n' % (read.qname, read.seq))
			out.close()
		    
                    

    def group(self, snvs):
        grouped_snvs = {}

        for snv in snvs:
	    seq = ""
	    if snv.snv_type in ('snv', 'ins', 'inv'):
		seq = snv.var_seq[:]
            coord = snv.coord()

	    snv_key = coord + "-" + seq
	    
            if not grouped_snvs.has_key(snv.snv_type):
                grouped_snvs[snv.snv_type] = {snv_key:[snv]}
            else:
                if not grouped_snvs[snv.snv_type].has_key(snv_key):
                    grouped_snvs[snv.snv_type][snv_key] = [snv]
                else:
                    grouped_snvs[snv.snv_type][snv_key].append(snv)

        return grouped_snvs
    
    def add_contig_reads(self, grouped_snvs):
	for type in ('snv', 'ins', 'del', 'inv'):
            if not grouped_snvs.has_key(type):
                continue
        
            for coord in grouped_snvs[type].keys():
		event_reads = 0
		for snv in grouped_snvs[type][coord]:
		    if snv.nreads_contig != 'na':
			event_reads += int(snv.nreads_contig)
		    
		for snv in grouped_snvs[type][coord]:
		    snv.nreads_event = event_reads

    def output_groups(self, grouped_snvs, output, debug=False):
        count1 = 1

        for type in ('snv', 'ins', 'del', 'inv'):
            if not grouped_snvs.has_key(type):
                continue
        
            for coord in grouped_snvs[type].keys():
                count2 = 1
                for snv in grouped_snvs[type][coord]:
                    if len(grouped_snvs[type][coord]) == 1:
                        count = count1
                    else:
                        count = "%d.%d" % (count1, count2)

                    output.write(str(count) + "\t" + snv.tab(debug=debug))
                
                    count2 += 1

                count1 += 1


    def report(self, outdir, post_filter=False):
        txt_file = outdir + "/snv.txt"
        gff_file = outdir + "/snv.gff"
        track_name = "snvs,indels"
	if not post_filter or not os.path.exists(txt_file):
	    self.output_txt(self.snvs, txt_file)
	    self.output_gff(self.snvs, gff_file, track_name)

	if post_filter:
	    self.post_filter(min_from_end=self.min_from_end)
	
	    if self.sample_type == 'genome':
		filtered_snvs = [snv for snv in self.snvs if not snv.artefact and snv.enough_coverage and not snv.too_close_to_end and snv.at_least_1_read_opposite ]
	    else:
		filtered_snvs = [snv for snv in self.snvs if not snv.artefact and snv.enough_coverage and not snv.too_close_to_end ]
	    
	    txt_file = outdir + "/snv_filtered.txt"
	    gff_file = outdir + "/snv_filtered.gff"
	    track_name = "snvs,indels(filtered)"
	    self.output_txt(filtered_snvs, txt_file)
	    self.output_gff(filtered_snvs, gff_file, track_name)
	
	    exon_snvs = [snv for snv in filtered_snvs if snv.exon and snv.nonsynon]
	    txt_file = outdir + "/snv_exons.txt"
	    self.output_txt(exon_snvs, txt_file)
        
	    debug_out = open(outdir + "/filter_debug.txt", 'w')
	    self.output_groups(self.grouped_snvs, debug_out, debug=True)
	    debug_out.close()
	
	    novel_snvs = [snv for snv in filtered_snvs if snv.dbsnp == '-']
	    txt_file = outdir + "/snv_filtered_novel.txt"
	    self.output_txt(novel_snvs, txt_file) 
	
	    novel_exon_snvs = [snv for snv in filtered_snvs if snv.exon and snv.nonsynon and snv.dbsnp == '-']
	    txt_file = outdir + "/snv_exons_novel.txt"
	    self.output_txt(novel_exon_snvs, txt_file) 

    def output_txt(self, snvs, outfile):
	grouped_snvs = self.group(snvs)
	
        out = open(outfile, 'w')
        out.write("\t".join(self.output_headers) + "\n")
        self.output_groups(grouped_snvs, out)
        out.close()
        
    def output_gff(self, snvs, outfile, track_name):
        snvs_snvs = [snv for snv in snvs if snv.snv_type and snv.snv_type.lower() == 'snv']
        snvs_indels = [snv for snv in snvs if snv.snv_type and not snv.snv_type.lower() == 'snv']
        out = open(outfile, 'w')
        if self.lib and len(self.lib) > 0:
            track_name = self.lib + " " + track_name
        out.write("track name=\"%s\" description=\"%s\" visibility=%d itemRgb=\"On\"\n" % (track_name, track_name, 3))
        for snv in snvs_snvs:
            out.write(snv.gff("exon"))
        for snv in snvs_indels:
            out.write(snv.gff("exon"))
        out.close()

    def post_filter(self, min_from_end=None):
        snvs_filtered = []
	
	kmer = re.compile(r'k(\d+):')
	
	self.filter_bad_region()

	for snv in self.snvs:
            if snv.nreads_event != 'na' and int(snv.nreads_event) >= int(self.min_reads):
		snv.enough_coverage = True
		
	    if snv.nreads_genome != 'na' and int(snv.nreads_genome) < int(self.min_reads_genome):
		snv.enough_coverage = False
            	    
	    if snv.gene and ('exon' in snv.gene or 'utr' in snv.gene):
		snv.exon = True
		
	    if snv.exon and not ':synon' in snv.gene:
		snv.nonsynon = True
		
	    #min_from_end only applies for non-bubbles (contig name ended with \d)
	    if re.match('\d', snv.var[-1]):
		if min_from_end == None:
		    m = kmer.search(snv.var)
		    if m and m.group:
			min_from_end_contig = int(m.group(1))
		   
			if int(snv.from_end) < min_from_end_contig:
			    snv.too_close_to_end = True
			
		elif int(snv.from_end) < int(min_from_end):
		    snv.too_close_to_end = True
		    
		    
    def filter_bad_region(self):
	#group by contig
	snvs_by_contig = {}
	for snv in self.snvs:
	    if snv.snv_type != 'snv' or len(snv.var_seq) != 1:
		continue
	    
	    if not snvs_by_contig.has_key(snv.var):
		snvs_by_contig[snv.var] = []
		
	    snvs_by_contig[snv.var].append(snv)
	    
	for contig, snvs in snvs_by_contig.iteritems():
	    snvs.sort(lambda x,y: int(x.var_start)-int(y.var_start))
	    
	    for i in range(len(snvs)):
		if i < len(snvs)-1:
		    diff = int(snvs[i+1].var_start) - int(snvs[i].var_start)
		else:
		    diff = 'na'
		
	    max_in_window = 10
	    window = 500
	    for i in range(len(snvs)-max_in_window+1):
		bad = [i]
		for j in range(i+1, len(snvs)):
		    if int(snvs[j].var_start) - int(snvs[i].var_start) < window:
			bad.append(j)
		    else:
			break
		    
		if len(bad) > max_in_window:
		    for idx in bad:
			snvs[idx].artefact = True
			
			
    def get_snvs(self, align, splice_motifs=None, refseq=None, cutoff=None):
        all_snvs = []
        sys.stderr.write("processing %s\n" % (align.query))
		
	if refseq:
	    if self.sample_type == 'transcriptome':
		align.correct_blocks(splice_motifs, refseq, align.contig.sequence)
		
	    all_snvs.extend(self.gap_snv(align, splice_motifs, refseq, align.contig.sequence, cutoff=cutoff))

	    if int(align.mismatch) > 0:
		all_snvs.extend(self.match_blocks(align, refseq, align.contig.sequence))

        return all_snvs
    
    def match_intron(self, ss, splice_motifs):
	if ss and (splice_motifs.has_key(ss.lower()) or splice_motifs.has_key(tools.reverse_complement(ss).lower())):
	    return True
	else:
	    return False
	
    def gap_snv(self, align, splice_motifs, refseq, query_seq, cutoff=None):
	"""
	identify insertions, deletions, inversion from gapped alignments
	"""
	
        snvs = []
	
        if self.debug:
            print align.query, align.blocks
            print align.query, align.query_blocks
            print align.splice_sites

	if self.sample_type == 'transcriptome' and not align.splice_sites:
	    return snvs
	    
	for i in range(len(align.blocks)-1):
	    if self.sample_type != 'transcriptome' or not self.match_intron(align.splice_sites[i], splice_motifs):
		if align.query_strand == '+':
		    qstart = align.query_blocks[i][1]+1
		    qend = align.query_blocks[i+1][0]-1
		    query = query_seq[qstart-1:qend]
		else:
		    qend = align.query_blocks[i][1]-1
		    qstart = align.query_blocks[i+1][0]+1
		    query = query_seq[qstart-1:qend]
		    query = tools.reverse_complement(query)
                    
		#target strand always + from blat
		tstart = align.blocks[i][1]+1
		tend = align.blocks[i+1][0]-1
		target = ''
		if tstart <= tend:
		    target = refseq.GetSequence(align.target, tstart, tend)
		
		#if code cannot extract sequence from reference, there must be a disagreement between alignment and reference - abort analysis
		if tend > tstart-1 and len(target) < 1:
		    sys.stderr.write("cannot extract reference sequence, abort: %s %s %s\n" % (align.target, tstart-1, tend))
		    sys.exit(100)

		snv_type = None
                
		if qstart > qend and (tend - tstart) >= 0:
		    size = tend - tstart + 1
		    if align.query_strand == '+':
			qstart = qend
		    else:
			qend = qstart
		    snv_type = "del"
		    
		elif tstart > tend and (qend - qstart) >= 0:
		    size = qend - qstart + 1
		    tstart = tend
		    snv_type = "ins"
		    
		else:
		    size = min(1, tend - tstart + 1)
		    snv_type = "indel"

		if size <= 0 or (cutoff and size > cutoff):
		    continue

		target = target.lower()
		query = query.lower()

		#print "gapsnv",snv_type, align.query, "%s:%d-%d" % (align.target, tstart, tend), qstart, qend, align.query_strand, target, query
		if not re.search('[^agtcATGC]', target) and not re.search('[^agtcATGC]', query):
		    if snv_type != 'indel':
			snv = SNV('blat', snv_type, align.target, tstart, tend, target, align.query_strand, align.query, qstart, qend, query)
			snvs.append(snv)
			
		    else:
			if len(query) == len(target) and (query[::-1].lower() == target.lower() or tools.reverse_complement(query).lower() == target.lower()):
			    #inversion must be longer than 1 base
			    if len(query) > 1:
				snv = SNV('blat', 'inv', align.target, tstart, tend, target, align.query_strand, align.query, qstart, qend, query)
				snvs.append(snv)
			    #1 bp gap in both query and target == snv
			    else:
				snv = SNV('blat', 'snv', align.target, tstart, tend, target, align.query_strand, align.query, qstart, qend, query)
				snvs.append(snv)
			else:
			    if align.query_strand == '+':
				qcoord = qstart
			    else:
				qcoord = qend
			    
			    snv = SNV('blat', 'del', align.target, tstart, tend, target, align.query_strand, align.query, qcoord, qcoord, query)
			    snvs.append(snv)
			    
			    tcoord = tstart
			    snv = SNV('blat', 'ins', align.target, tcoord, tcoord, target, align.query_strand, align.query, qstart, qend, query)
			    snvs.append(snv)
			                    
        return snvs

    def parse_results(self, snv_file, select_types=None):
	"""
	parse results from text output that can be updated with annotation, read-support, etc
	"""
	
	names = SNVCaller.output_headers

	field_name_conversion = {
	    'type': 'snv_type',
	    'chr': 'ref',
	    'chr_start': 'ref_start',
	    'chr_end': 'ref_end',
	    'ctg': 'var',
	    'ctg_len': 'var_len',
	    'ctg_start': 'var_start',
            'ctg_end': 'var_end',
	    'len': 'snv_len',
            'ref': 'ref_seq',
	    'alt': 'var_seq',
	    'event_reads': 'nreads_event',
            'contig_reads': 'nreads_contig',
	    'genome_reads': 'nreads_genome',
            'gene': 'gene',
	    'from_end': 'from_end',
	    'ctg_strand': 'query_strand',
	    }

        for line in open(snv_file, 'r'):
            cols = line.rstrip('\n').split('\t')
	    
	    if cols[0] == 'id':
		continue

            attributes = {}
            for i in range(1, len(cols)):
                name = names[i]
		value = cols[i]
		
		if field_name_conversion.has_key(name):
		    name = field_name_conversion[name]
		    
		if name in ('expansion', 'from_end'):
		    value = int(value)
		elif name == 'confirm_contig_region':
		    value = value.split('-')
		    value[0] = int(value[0])
		    value[1] = int(value[1])
		elif name == 'at_least_1_read_opposite':
		    if value == 'true':
			value = True
		    else:
			value = False
		    
                attributes[name] = value
	    
	    if select_types and not attributes['snv_type'] in select_types:
		continue
	    
            snv = SNV(method='blat')
            tools.set_attrs(snv, attributes)
            self.snvs.append(snv)
	    
    def parse_results_dir(self, path):
	"""
	parse results into single file given a directory of output directories
	"""
	
	output_dirs = sorted(glob.glob(os.path.join(path, '*')))
		
	output_files = []
	num_output_dirs = 0
	missing_dirs = []
	for job_num in range(1, len(output_dirs)+1):
	    cluster_outdir = "%s/%s" % (options.output_file, job_num)
	    if os.path.isdir(cluster_outdir):
		num_output_dirs += 1
		    
		snv_file = cluster_outdir + '/snv.txt'		    
		if os.path.exists(snv_file):
		    output_files.append(snv_file)
		else:
		    missing_dirs.append(str(job_num))
		    sys.stdout.write("%s does not have output\n" % (cluster_outdir))
			
	sys.stdout.write("output dirs:%s output files:%s\n" % (num_output_dirs, len(output_files)))
		
	if num_output_dirs == len(output_files):
	    concat_outfile = options.outdir + "/snv_concat.txt"
	    print concat_outfile
	    concat_out = open(concat_outfile, 'w')
		    
	    #just for check if header is to be written
	    count = 0
	    for output_file in output_files:
		infile = open(output_file, 'r')
		if count == 0:
		    concat_out.writelines(open(output_file, 'r').readlines())
		else:
		    concat_out.writelines(open(output_file, 'r').readlines()[1:])
		infile.close()
		    
		self.parse_results(output_file)
		count += 1
			
	    concat_out.close()
	else:
	    sys.stdout.write("missing (%s):%s\n" % (len(missing_dirs), ','.join(missing_dirs)))
	    
    def match_blocks(self, align, refseq, query_seq):
        snvs = []
        
        for i in range(len(align.blocks)):
	    if align.query_strand == '+':
                qseq = query_seq[int(align.query_blocks[i][0])-1:int(align.query_blocks[i][1])]
            else:
                qseq = tools.reverse_complement(query_seq[int(align.query_blocks[i][1]-1):int(align.query_blocks[i][0])])
	    tseq = refseq.GetSequence(align.target, int(align.blocks[i][0]), int(align.blocks[i][1]))
	    
            mismatches = self.find_mismatches(qseq, tseq)
	    
            for pos, change in mismatches.iteritems():
                tpos = int(align.blocks[i][0]) + pos
                if int(align.query_blocks[i][0]) < int(align.query_blocks[i][1]):
                    qpos = int(align.query_blocks[i][0]) + pos
                else:
                    qpos = int(align.query_blocks[i][0]) - pos
		snv = SNV('blat', 'snv', align.target, tpos, tpos, change[0], align.query_strand, align.query, qpos, qpos, change[1])
                snvs.append(snv)
		
        return snvs

    def find_mismatches(self, qseq, tseq):
	"""
	given query and target sequence of same length report substitutions
	"""
	
        pos = {}
	bases = ['a','g','t','c']
        
        if len(qseq) == len(tseq):
            for i in range(len(qseq)):
                if qseq[i].lower() != tseq[i].lower() and qseq[i].lower() in bases and tseq[i].lower() in bases:
                    pos[i] = [tseq[i].lower(), qseq[i].lower()]
        return pos
    
    def curate_inversion(self, refseq):
	"""
	create inversion events from pre-existing data
	i.e. from pairs of ins and del events at same genome coordinate and of same length but with inverted sequence
	"""
	
	ins_indices = [i for i in range(len(self.snvs)) if self.snvs[i].snv_type == 'ins']
	del_indices = [i for i in range(len(self.snvs)) if self.snvs[i].snv_type == 'del']
	del_by_coord = dict((self.snvs[i].ref + ' ' + self.snvs[i].ref_start, i) for i in del_indices)
	
	indices_delete = []
	for ii in ins_indices:
	    if int(self.snvs[ii].snv_len) < 2:
		continue
	    
	    key = self.snvs[ii].ref + ' ' + self.snvs[ii].ref_start
	    
	    if del_by_coord.has_key(key):
		di = del_by_coord[key]
		
		if self.snvs[ii].snv_len == self.snvs[di].snv_len:
		    ins_seq = self.snvs[ii].var_seq
		    
		    if '.' in self.snvs[di].ref_seq:
			del_seq = refseq.GetSequence(self.snvs[di].ref, int(self.snvs[di].ref_start), int(self.snvs[di].ref_end))
		    else:
			del_seq = self.snvs[di].ref_seq
		    
		    if ins_seq[::-1].lower() == del_seq.lower() or tools.reverse_complement(ins_seq).lower() == del_seq.lower():
			print 'ins',self.snvs[ii].ref, self.snvs[ii].ref_start, self.snvs[ii].var_seq, self.snvs[ii].snv_len, ins_seq
			print 'del',self.snvs[di].ref, self.snvs[di].ref_start, self.snvs[di].ref_seq, self.snvs[di].snv_len, del_seq
			print '---'
			
			indices_delete.append(ii)
			indices_delete.append(di)
	
			snv = SNV('blat', 'inv', self.snvs[di].ref, self.snvs[di].ref_start, self.snvs[di].ref_end, del_seq, self.snvs[di].query_strand, self.snvs[ii].var, self.snvs[ii].var_start, self.snvs[ii].var_end, ins_seq, var_len=self.snvs[ii].var_len)
			snv.from_end = self.snvs[ii].from_end
			self.snvs.append(snv)
			
	if len(indices_delete) > 0:
	    indices_delete.sort(lambda x,y: y-x)
	    
	    for i in indices_delete:
		del self.snvs[i]
		
    
def create_jobs_file(align_dir, outdir, num):
    """
    creates cluster jobs file
    """
    
    cluster_dir = outdir + "/cluster"
    if not os.path.isdir(cluster_dir):
	os.makedirs(cluster_dir)
	
    jobs_file = cluster_dir + "/jobs"
    sys.stdout.write("creating job file:%s\n" % (num))
    jobs = open(jobs_file, 'w')
    
    script = os.path.abspath(sys.argv[0])
    source = PACKAGE_DIR + "/setup"

    for i in range(1,num+1):
	outdir = cluster_dir + "/" + str(i)
	if not os.path.isdir(outdir):
	    os.mkdir(outdir)
	
	align_file = "%s/seq.%s.psl" % (align_dir, i)
	contigs_file = "%s/seq.%s.fa" % (options.contigs_file, i)
	if options.contigs_bam and options.gene_model:
	    job = "source %s;%s %s -a %s -o %s -g %s -c %s -C %s -m %s" % (source, PYTHON, script, align_file, outdir, options.genome, contigs_file, options.contigs_bam, options.gene_model)
	elif options.contigs_bam:
	    job = "source %s;%s %s -a %s -o %s -g %s -c %s -C %s" % (source, PYTHON, script, align_file, outdir, options.genome, contigs_file, options.contigs_bam)
	elif options.gene_model:
	    job = "source %s;%s %s -a %s -o %s -g %s -c %s -m %s" % (source, PYTHON, script, align_file, outdir, options.genome, contigs_file, options.gene_model)
	else:
	    job = "source %s;%s %s -a %s -o %s -g %s -c %s" % (source, PYTHON, script, align_file, outdir, options.genome, contigs_file)
	    
	if options.lib:
	    job += " -L " + options.lib
	if options.add_contig_reads:
	    job += " -S"
	if options.min_from_end:
	    job += " -E " + str(options.min_from_end)
	if options.bubble2contig:
	    job += " -B " + options.bubble2contig
	if options.sample_type:
	    job += " -T " + options.sample_type
	if options.olap_annot:
	    job += " -O"
	if options.filter:
	    job += " -X"
	if options.genome_bam:
	    job += " -G " + options.genome_bam
	    
	jobs.write(job + '\n')
    jobs.close()
    
    return jobs_file


if __name__ == '__main__':
    usage = "Usage: %prog [options]"
    parser = OptionParser(usage=usage, version="%prog " + VERSION)
    parser.add_option("-a", "--align_file", dest="align_file", help="alignment file")
    parser.add_option("-t", "--is_track", dest="is_track", help="is a track file", action="store_true", default=False)
    parser.add_option("-c", "--contigs_file", dest="contigs_file", help="contigs file")
    parser.add_option("-g", "--genome", dest="genome", help="genome")
    parser.add_option("-o", "--outdir", dest="outdir", help="outdir")
    parser.add_option("-i", "--indel_size", dest="indel_size", help="indel upper size limit for blat", type='int')
    parser.add_option("-L", "--lib", dest="lib", help="library (for reading contig bam file)")
    parser.add_option("-G", "--genome_bam", dest="genome_bam", help="genomic bam file")
    parser.add_option("-C", "--contigs_bam", dest="contigs_bam", help="contigs bam file")
    parser.add_option("-m", "--gene_model", dest="gene_model", help="gene model used for annotation e.g. k where k=known_genes, e=ensembl, r=refseq")
    parser.add_option("-r", "--min_reads_contigs", dest="min_reads_contigs", help="minimum number of read support from contig bam. Default=3", default=3, type='int')
    parser.add_option("-n", "--min_reads_genome", dest="min_reads_genome", help="minimum number of read support from genome bam. Default=1", default=1, type='int')
    parser.add_option("-b", "--read_buffer", dest="read_buffer", help="minimum number of bases from read end", default=8, type='int')
    parser.add_option("-d", "--debug", dest="debug", help="debug", action="store_true", default=False)
    parser.add_option("-f", "--output_file", dest="output_file", help="output file")
    parser.add_option("-k", "--cluster", dest="cluster", help="cluster")
    parser.add_option("-S", "--add_contig_reads", dest="add_contig_reads", action="store_true", default=False)
    parser.add_option("-E", "--contig_end", dest="min_from_end", help="minimum distance from contig end", type='int')
    parser.add_option("-B", "--bubble2contig", dest="bubble2contig", help="bubble to contig alignment")
    parser.add_option("-T", "--type", dest="sample_type", help="sample type: transcriptome(default), genome, exome", default='transcriptome')
    parser.add_option("-M", "--match_percent", dest="match_percent", help="minimum match percent for screening alignments. Default:95", type='float', default=95.0)
    parser.add_option("-I", "--identity", dest="identity", help="minimum identity for screening alignments Default:95", type='float', default=95.0)
    parser.add_option("-v", "--curate_inversion", dest="curate_inversion", help="merge ins and del events into inv", action="store_true", default=False)
    parser.add_option("-O", "--olap_annot", dest="olap_annot", help="overlaps repeats and dbsnp when parsing events", action="store_true", default=False)
    parser.add_option("-X", "--filter", dest="filter", help="post-filter", action="store_true", default=False)
    parser.add_option("-R", "--output_read_support", dest="output_read_support", help="output read support", action="store_true", default=False)
    
    (options, args) = parser.parse_args()
    
    tools.output_log(VERSION, sys.argv, [vars(options)], options.outdir)

    snv_caller = SNVCaller()
    snv_caller.sample_type = options.sample_type
    snv_caller.genome = options.genome
    snv_caller.lib = options.lib
    snv_caller.min_reads = options.min_reads_contigs
    snv_caller.min_reads_genome = options.min_reads_genome
    snv_caller.debug = options.debug
    snv_caller.min_from_end = options.min_from_end
    
    if options.outdir and not os.path.isdir(options.outdir):
	os.makedirs(options.outdir)
    
    bm = None
    if options.bubble2contig:
	bm = BubbleMapper(options.bubble2contig)
	bm.map_to_contigs()
	snv_caller.bubble_mapping = bm
	
    refseq = None
    if options.genome:
	refseq = tools.get_refseq_from_2bit(options.genome)
	snv_caller.get_annotation_files()
		
	if options.gene_model:
	    snv_caller.feature_finder = FeatureFinder(options.genome, options.gene_model)
	
    #extract alignments (either a single track or alignment file(s)
    if options.align_file:	
	#if given directory of align files, either run code or create jobs file(given cluster)
	if os.path.isdir(options.align_file):
	    align_files = sorted(glob.glob(os.path.join(options.align_file, '*.*')))
		
	    if not options.cluster:		
		for i in range(len(align_files)):
		    snv_caller.align_file = "%s/seq.%s.psl" % (options.align_file, i+1)
		    
		    if os.path.isdir(options.contigs_file):
			snv_caller.contigs_file = "%s/seq.%s.fa" % (options.contigs_file, i+1)
		    else:
			snv_caller.contigs_file = options.contigs_file
			
		    sys.stdout.write("processing alignment file #%d\n" % (i+1))
		    snv_caller.extract(refseq=refseq, cutoff=options.indel_size, model=options.gene_model, no_group=True, match_percent=options.match_percent, identity=options.identity)
	    	    
	    else:
		jobs_file = create_jobs_file(options.align_file, options.outdir, len(align_files))
	
	else:
	    snv_caller.align_file = options.align_file
	    snv_caller.contigs_file = options.contigs_file
	    snv_caller.is_track = options.is_track
	    snv_caller.extract(refseq=refseq, cutoff=options.indel_size, model=options.gene_model, match_percent=options.match_percent, identity=options.identity)
		
    #parse result files
    elif options.output_file:	
	if os.path.isdir(options.output_file):
	    snv_caller.parse_results_dir(options.output_file)
	else:
	    snv_caller.parse_results(options.output_file)  
	   
	if options.curate_inversion:
	    snv_caller.curate_inversion(refseq)
	    for snv in snv_caller.snvs:
		if snv.snv_type == 'inv':
		    snv.gene = snv_caller.feature_finder.get_feature(' '.join((snv.ref, str(snv.ref_start), str(snv.ref_end))), refseq=refseq, variant=snv.var_seq, change=snv.snv_type)
	        
    
    #processing regardless parsing alignment file or result file
    #1. group events
    #2. if genome and annotation provided, annotate feature
    #3. if bam file provide, update read support
    #4. add contig reads in groups if desired
    #5. report
    if not (options.align_file and options.cluster) and snv_caller:
	snv_caller.grouped_snvs = snv_caller.group(snv_caller.snvs)
	
	#genome and gene model given, indicate annotating feature
	if snv_caller.feature_finder:
	    snv_caller.annotate_genes(refseq)
	
        genome_bam = contigs_bam = None
        if options.genome_bam:
            genome_bam = BAM(options.genome_bam)
        if options.contigs_bam:
            contigs_bam = BAM(options.contigs_bam, min_mapq=0)
           
	#bam file given, indicate to calculate read support
	if contigs_bam or genome_bam:
	    reads_support_dir = options.outdir + "/reads_support"
	    if os.path.isdir(reads_support_dir):
		os.system("rm -Rf " + reads_support_dir)
		
	    if options.output_read_support:
		os.mkdir(reads_support_dir)
	    else:
		reads_support_dir = None
	    
	    if options.curate_inversion:
		snv_caller.add_support(genome_bam=genome_bam, contigs_bam=contigs_bam, lib=options.lib, outdir=reads_support_dir, from_end=options.read_buffer, specific_type='inv', refseq=refseq)
	    else:
		snv_caller.add_support(genome_bam=genome_bam, contigs_bam=contigs_bam, lib=options.lib, outdir=reads_support_dir, from_end=options.read_buffer, refseq=refseq)

	if options.add_contig_reads:
	    snv_caller.add_contig_reads(snv_caller.grouped_snvs)
	    
	if options.olap_annot:
	    snv_caller.overlap_repeats()
	    snv_caller.overlap_dbsnp(refseq)
		
        snv_caller.report(options.outdir, post_filter=options.filter)
