from align_parsers import exonerate,blat
import sys, os, re, glob
from optparse import OptionParser
import assembly
from track import Track

#generic alignment parsing functions not specific to particular aligner

def parse(file, aligner, filters=None, splice_motif_file=None, noline=None, refseq=None):
    print file, aligner

    #do filtering at the parsing level will reduce creating unwanted objects
    #do sorting here because all alignments are the same regardless of aligners
    aligns = {
        'exonerate': exonerate.parse,
        'blat': blat.parse,
        }[aligner](file, filters, splice_motif_file=splice_motif_file, noline=noline, refseq=refseq)

    print "total:", len(aligns)
        
    return aligns

def get_targets(file_name):
    "get custom targets"

    targets = []
    f = open(file_name, 'r')
    for line in f.readlines():
        if re.search('\d', line):
            (label, target) = re.split('\t+', line.rstrip('\n').lstrip(' '))
            target = re.sub('(chr|,)', '', target)
            targets.append(target)
    f.close()

    return targets


def output_aligns(outfile, aligns, track_name=None, contigs=None, append=False, header=True, genome=None):
    ext = os.path.splitext(outfile)[1]
    if append:
        out = open(outfile, 'a')
    else:
        out = open(outfile, 'w')

    track_desc = track_name
    #track_name = track_desc

    #alignments - output track
    if aligns:
        if track_name and genome:
            Track.ucsc_targets(genome, aligns)
        
        if header and track_name:
            out.write("track name=\"%s\" description=\"%s\" visibility=%d itemRgb=\"On\"\n" % (track_name, track_desc, 3))

        #map contig to contig name
        contig_dict = None
        if contigs:
            contig_dict = dict((contig.num, contig) for contig in contigs)
            
        for align in aligns:
            #this is for adding coverage info
            contig = align.query
            if ' ' in contig:
                contig = align.query.split(" ")[0]
            
            if contig_dict and contig_dict.has_key(contig):
                align.contig = contig_dict[contig]
                                
            if ext == ".gff":
                out.write(align.gff("exon"))

            elif ext == ".psl":
                out.write(align.psl())

            elif ext == ".exon":
                out.write(align.exon())
                
    out.close()


def get_contigs(fasta, kmer=None, lib=None):
    sys.stderr.write("extracting contigs...")

    if kmer:
        ass = assembly.Assembly(lib, k=kmer)
    else:
        ass = assembly.Assembly(lib, k=1)
        
    ass.fasta = fasta
    contigs = ass.get_contigs()

    for contig in contigs:
        if ":" in contig.num:
            k, ctg = contig.num.split(':')
            contig.k = k.lstrip('k')
        elif kmer:
            contig.k = kmer

    sys.stderr.write("done\n")
    return contigs


def batch(aligner, indir, ext=None):
    if not ext:
        if aligner == 'blat':
            ext = '.psl'
        elif aligner == 'exonerate':
            ext = '.exon'

    inputs = glob.glob(os.path.join(indir + '/*.'+'[0-9]*' + ext))
    return sorted(inputs)


if __name__ == '__main__':
    usage = "Usage: %prog file aligner [options]"
    parser = OptionParser(usage=usage)
    parser.add_option("-n", "--bestn", dest="bestn", help="best hit filter")
    parser.add_option("-i", "--identity", dest="identity", help="identity filter")
    parser.add_option("-m", "--match", dest="match_percent", help="match length percentage")
    parser.add_option("-u", "--unique", dest="unique", help="unique", action="store_true", default=False)
    parser.add_option("-q", "--query_len", dest="query_len", help="query length filter")
    parser.add_option("-t", "--target", dest="target", help="target e.g. chr1:35422000-35425800, chr13, or file containing targets")
    parser.add_option("-o", "--out", dest="out", help="output file, with proper file extension e.g. x.gff, x.psl, x.snv, or x.gff,x.snv")
    parser.add_option("-k", "--track", dest="track_name", help="track name e.g. \"HS1202 k45\"")
    parser.add_option("-f", "--fasta", dest="fasta", help="fasta file, for extracting contig info")
    parser.add_option("-K", "--k", dest="k", help="kmer value, if not present in contig name")
    parser.add_option("-d", "--batch", dest="batch", help="batch mode", action="store_true", default=False)
    parser.add_option("-g", "--genome", dest="genome", help="genome for UCSC chromosome conversion")
    (options, args) = parser.parse_args()
    
    filters = {}
    if len(args) == 2:
        if options.bestn:
            filters['bestn'] = int(options.bestn)
        if options.identity:
            filters['identity'] = float(options.identity)
        if options.match_percent:
            filters['match'] = float(options.match_percent)
        if options.unique:
            filters['unique'] = True
        if options.query_len:
            filters['qlen'] = int(options.query_len)
        if options.target:
            if os.path.exists(options.target):
                filters['target'] = get_targets(options.target)
            else:
                filters['target'] = [options.target]

        #get contig/coverage
        contigs = None
        if options.fasta:
            contigs = get_contigs(options.fasta, options.k)

        #inputs
        inputs = []
        if options.batch:
            inputs = batch(args[1], args[0])
        else:
            inputs.append(args[0])

        if options.batch:
            if options.out and os.path.exists(options.out):
                os.remove(options.out)
                
        for i in range(len(inputs)):
            infile = inputs[i]
            #for deciding if header needs to be output
            first = False
            if i == 0:
                first = True

            #this is specifically for outputting exonerate original alignments
            if options.out and os.path.splitext(options.out)[1] == ".exon":
                aligns = parse(infile, args[1], filters, noline=False)
            else:
                aligns = parse(infile, args[1], filters)
            
            if options.out:
                output_aligns(options.out, aligns, options.track_name, contigs=contigs, append=options.batch, header=first, genome=options.genome)

    else:
        parser.error("incorrect number of arguments")
