import os, sys, re
from optparse import OptionParser
import subprocess
from utilities import cfg

number_pattern = re.compile('(\d+)')
script_dir = os.path.dirname(os.path.abspath(sys.argv[0]))
fachain = "fachain"
BLAT = 'blat'

abyss_versions_1 = ['1.1.2', '1.2.0', '1.2.1', '1.2.2'] #versions that have proper *-4.fa
abyss_versions_2 = ['1.2.3', '1.2.4', '1.2.5'] #versions that need some work on *-4.fa
abyss_all = abyss_versions_1 + abyss_versions_2 #all acceptable versions
    
def get_contig_num(num_orient):
    return number_pattern.search(num_orient).group(1)

class Assembly:

    def __init__(self, lib, abyss=None, k=None, loc=None, fasta=None):
        self.lib = lib
        self.k = k
        self.loc = loc
        self.fasta = fasta
        self.junctions = []
        self.pet = None
        self.abyss = abyss
        
        if self.loc:
            self.pet = loc + "/" + lib + "-contigs.fa"
            self.overlap = loc + "/" + lib + "-3-overlap.fa"
            self.adj_bubbles = loc + "/" + lib + "-4.adj"
            self.adj_islands = loc + "/" + lib + "-4.adj"
            self.indel = loc + "/" + lib + "-indel.fa"

    def get_contigs(self, ids=[], length=0, len_range=[], sequence=False):
        "get contig data"
        
        contigs = []

        fasta = None
        if self.fasta:
            fasta = self.fasta
        elif self.pet:
            fasta = self.pet

        ids_dict = None
        if ids:
            ids_dict = dict((c, True) for c in ids)
            
        if fasta and os.path.exists(fasta):
            seq = ""
            last_contig = None
            for line in open(fasta, 'r'):
                #header
                if line[0] == '>':
                    if last_contig:
                        if sequence and len(seq) > 0:
                            last_contig.sequence = seq
                            if not last_contig.length:
                                last_contig.length = len(seq)
                                
                        contigs.append(last_contig)
                            
                    seq = ""
                    
                    cols = line.rstrip('\n').lstrip('>').split(' ')

                    good = True
                    if ids and not ids_dict.has_key(cols[0]):
                        good = False

                    if good:
                        if length and int(length) > 0 and int(cols[1]) != int(length):
                            good = False

                        if len_range:
                            #proper range given
                            if int(len_range[1]) > int(len_range[0]):
                                if (int(cols[1]) < int(len_range[0])) or (int(cols[1]) > int(len_range[1])):
                                    good = False
                            #first number less than second number, indicates check minimum
                            elif int(cols[1]) < int(len_range[0]):
                                good = False
                    if good:
                        size, coverage, children = None, None, None
                        if len(cols) >= 3:
                            #length, coverage, children = int(cols[1]), int(cols[2]), cols[3]
                            if cols[1] and not re.search('\D', cols[1]):
                                size = int(cols[1])
                            if cols[2] and not re.search('\D', cols[2]):
                                coverage = int(cols[2])
                        if len(cols) > 3 and cols[3]:
                            children = cols[3]
                        last_contig = Contig(cols[0], length=size, coverage=coverage, children=children)
                    else:
                        last_contig = None
                    
                elif sequence and line[0] != ' ':
                    seq += line.rstrip('\n')

            if last_contig:
                if sequence and len(seq) > 0:
                    last_contig.sequence = seq
                    if not last_contig.length:
                        last_contig.length = len(seq)
                        
                contigs.append(last_contig)

            seq = ""
            

        return contigs

    def contigs2dict(self, contigs):
        contig_dict = {}
        for contig in contigs:
            contig_dict[contig.num] = {}
            contig_dict[contig.num]['sequence'] = contig.sequence
            contig_dict[contig.num]['coverage'] = contig.coverage
            contig_dict[contig.num]['length'] = contig.length

        return contig_dict

    def output(self, out_file):
        out = open(out_file, 'w')

        for contig in self.contigs:
            out.write(contig.fasta())

        out.close()

    def get_junctions(self, cutoff=None, outdir=None):
        "extracts all 1-in-1-out junctions beteen k and 2k-2"

        #get all 1 in 1 out
        junctions = {}
        for line in open(self.adj_bubbles, 'r'):
            middle, right, left = line.rstrip('\n').split(';')
            middle, size = middle.rstrip().split(' ')[0:2]
            left = [c for c in left.lstrip().rstrip().split(' ') if re.search('[0-9]', c)]
            right = [c for c in right.lstrip().rstrip().split(' ') if re.search('[0-9]', c)]
            
            if len(left) == 1 and len(right) == 1:
                junctions[middle] = True

        #only keep the 1 in 1 out that are single end contigs
        se_junctions = []
        for contig in self.contigs:
            if not contig.is_pet() and junctions.has_key(contig.num):
                se_junctions.append(contig.num)

        #extend junctions
        jn_fa = self.extend_junctions(se_junctions, outdir=outdir)
        self.fasta = jn_fa
        jn_contigs = self.get_contigs(sequence=True, ids=se_junctions)

        #sets size range
        if not cutoff:
            min_size = int(self.k)
        else:
            min_size = int(cutoff)

        remove = []
        for i in range(len(jn_contigs)):
            contig = jn_contigs[i]
            
            if int(contig.length) < min_size:
                remove.append(i)

            elif not contig.children:
                remove.append(i)

        remove.reverse()
        for i in remove:
            del jn_contigs[i]

        self.junctions = jn_contigs
        
        sys.stderr.write("junctions(%d):%d\n" % (min_size, len(self.junctions)))
        os.system('rm -Rf ' + os.path.split(jn_fa)[0])
                
    def extend_junctions(self, junctions, outdir=None):
        tmpdir = '-'.join((self.lib, self.k, 'jn'))
        if outdir:
            tmpdir = outdir + '/' + tmpdir
        
        if os.path.isdir(tmpdir):
            os.system('rm -Rf ' + tmpdir)
        os.mkdir(tmpdir, 2775)
        os.chmod(tmpdir, 02775)

        #print tmpdir
        jn_list = tmpdir + '/junctions.txt'
        out = open(jn_list, 'w')
        for jn in junctions:
            out.write(jn + '\n')
        out.close()

        #if os.path.exists(self.overlap):
        #    input_fa = self.loc + "/" + self.lib + "-4.fa"
        if self.abyss in abyss_versions_1: #abyss 1.2.2 and earlier
            input_fa = self.loc + "/" + self.lib + "-4.fa"
        elif self.abyss == "1.2.3": #abyss 1.2.3, has overlap.fa but no -4.fa
            input_fa = tmpdir + "/" + self.lib + "-4.fa"
            dash3_fa = self.loc + "/" + self.lib + "-3.fa"
            if not os.path.exists(dash3_fa) or not os.path.exists(self.overlap):
                sys.exit(dash3_fa + " or " + self.overlap + " file not found in ABySS 1.2.3 output")
            cmd = 'cat ' + dash3_fa + ' ' + self.overlap + ' >' + input_fa
            os.system(cmd)
        else: #abyss 1.2.4 and 1.2.5
            input_fa = tmpdir + "/" + self.lib + "-34.fa"

            if os.path.exists(input_fa):
                os.remove(input_fa)
            for i in range(3,5):
                source = self.loc + "/" + self.lib + "-" + str(i) + ".fa"
                cmd = 'cat ' + source + ' >> ' + input_fa
                #print cmd
                os.system(cmd)
            
        final_fa = 'extended-junction-contigs.fa'

        #simplegraph
        simplegraph = "SimpleGraph"
        args = [simplegraph]
        args.append('-k' + str(self.k))
        args.append('-o' + tmpdir + '/junction.path')
        args.append('-c' + tmpdir + '/junction.cycpath')
        args.append('-m -vv -j1')
        args.append(self.loc + "/" + self.lib + '-4.adj')
        args.append(jn_list)
        args.append('>' + tmpdir + '/junctions.path.out 2>&1') 
        cmd = ' '.join(args)
        #print cmd
        os.system(cmd)

        #mergecontigs
        mergecontigs = "MergeContigs"
        args = [mergecontigs]
        args.append('-k' + str(self.k))
        args.append('-v')
        args.append('-o' + tmpdir + '/' + final_fa)
        args.append(input_fa)
        args.append(tmpdir + '/junction.path')
        args.append('>' + tmpdir + '/junctions.mergecontigs.out 2>&1') 
        cmd = ' '.join(args)
        #print cmd
        os.system(cmd)

        return tmpdir + '/' + final_fa

    def get_indels(self, outdir=None):
        indels = []
        
        if os.path.exists(self.indel):
            for line in open(self.indel, 'r'):
                if re.match('^>', line):
                    contig_id = line.split(' ')[0][1:]
                    indels.append(contig_id)

        if indels:
            indel_fa = self.extend_indels(indels, outdir=outdir)
            self.fasta = indel_fa
            indel_contigs = self.get_contigs(sequence=True, ids=indels)
            self.indels = indel_contigs
            
            os.system('rm -Rf ' + os.path.split(indel_fa)[0])
        else:
            self.indels = []
        
        sys.stderr.write("indels:%d\n" % (len(self.indels)))
        
    def extend_indels(self, indels, outdir=None):
        tmpdir = '-'.join((self.lib, self.k, 'indel'))
        if outdir:
            tmpdir = outdir + '/' + tmpdir
        
        if os.path.isdir(tmpdir):
            os.system('rm -Rf ' + tmpdir)
        os.mkdir(tmpdir)
        os.chmod(tmpdir, 02775)

        #print tmpdir
        indel_list = tmpdir + '/indels.txt'
        out = open(indel_list, 'w')
        for indel in indels:
            out.write(indel + '\n')
        out.close()

        input_fa = self.loc + "/" + self.lib + "-1.fa"
        final_fa = 'extended-indels.fa'

        #simplegraph
        simplegraph = "SimpleGraph"
        args = [simplegraph]
        args.append('-k' + str(self.k))
        args.append('-o' + tmpdir + '/indel.path')
        args.append('-c' + tmpdir + '/indel.cycpath')
        args.append('-m -vv -j1')
        args.append(self.loc + "/" + self.lib + '-1.adj')
        args.append(indel_list)
        args.append('>' + tmpdir + '/indels.path.out 2>&1') 
        cmd = ' '.join(args)
        #print cmd
        os.system(cmd)

        #mergecontigs
        mergecontigs = "MergeContigs"
        args = [mergecontigs]
        args.append('-k' + str(self.k))
        args.append('-v')
        args.append('-o' + tmpdir + '/' + final_fa)
        args.append(input_fa)
        args.append(tmpdir + '/indel.path')
        args.append('>' + tmpdir + '/indels.mergecontigs.out 2>&1') 
        cmd = ' '.join(args)
        #print cmd
        os.system(cmd)

        return tmpdir + '/' + final_fa
        
    def remove_redundant(self):
        "remove set contigs used in contructing junctions"
        contigs_remove = {}
        junctions_remove = {}

        #get rid of se contigs used for junctions
        jn_se = {}
        for jn in self.junctions:
            for se in [get_contig_num(c) for c in jn.children.split(',')]:
                jn_se[se] = True

        for i in range(len(self.contigs)):
            if not self.contigs[i].is_pet() and jn_se.has_key(self.contigs[i].num):
                contigs_remove[self.contigs[i].num] = True

        del jn_se

        indel_se = {}
        for indel in self.indels:
            if indel.children:
                for se in [get_contig_num(c) for c in indel.children.split(',')]:
                    indel_se[se] = True

        for i in range(len(self.contigs)):
            if not self.contigs[i].is_pet() and indel_se.has_key(self.contigs[i].num):
                contigs_remove[self.contigs[i].num] = True

        del indel_se

        #get rid of pe contigs used for junctions
        jn_children = {}
        for i in range(len(self.junctions)):
            children = [str(get_contig_num(c)) for c in self.junctions[i].children.split(',')]

            forward = ','.join(children)
            children.reverse()
            reverse = ','.join(children)
            for se in children:
                jn_children[se] = [forward, reverse]

        indel_children = {}
        for i in range(len(self.indels)):
            if not self.indels[i].children:
                continue
            children = [str(get_contig_num(c)) for c in self.indels[i].children.split(',')]

            forward = ','.join(children)
            children.reverse()
            reverse = ','.join(children)
            for se in children:
                indel_children[se] = [forward, reverse]

        for i in range(len(self.contigs)):
            if self.contigs[i].is_pet():
                children = [str(get_contig_num(c)) for c in self.contigs[i].children.split(',')]
                forward = ','.join(children)

                for se in children:
                    if jn_children.has_key(se):
                        if forward in jn_children[se][0] or forward in jn_children[se][1]:
                            contigs_remove[self.contigs[i].num] = True
                            break
                        
                    if indel_children.has_key(se):
                        if forward in indel_children[se][0] or forward in indel_children[se][1]:
                            contigs_remove[self.contigs[i].num] = True
                            break


        #get rid of junctions that are subsets of other junctions
        for i in range(len(self.junctions)):
            children = [str(get_contig_num(c)) for c in self.junctions[i].children.split(',')]
            forward = ','.join(children)

            for se in children:
                if jn_children.has_key(se):
                    if forward == jn_children[se][0] or forward == jn_children[se][1]:
                        continue
                    
                    if forward in jn_children[se][0] or forward in jn_children[se][1]:
                        junctions_remove[self.junctions[i].num] = True
                        break
                
        return contigs_remove, junctions_remove

    def get_seqs(self, contigs):
        "get sequences given list of contig objects"
        
        ids = dict((c.num, c) for c in contigs)

        fasta = None
        if self.pet:
            fasta = self.pet
        elif self.fasta:
            fasta = self.fasta

        if fasta and os.path.exists(fasta):
            seq = ""
            contig = None

            for line in open(fasta, 'r'):
                if line[0] == '>':
                    if contig and len(seq) > 0:
                        contig.sequence = seq
                        if not contig.length:
                            contig.length = len(seq)
                            
                    seq = ""
                    
                    cols = line.rstrip('\n').lstrip('>').split(' ')

                    if ids.has_key(cols[0]):
                        contig = ids[cols[0]]
                    else:
                        contig = None
                        
                elif line[0] != ' ':
                    seq += line.rstrip('\n')
                    
            if contig and len(seq) > 0:
                contig.sequence = seq
                if not contig.length:
                    contig.length = len(seq)
    
    def get_seq(self, contig):
        process = subprocess.Popen(["grep", "-1", contig, self.fasta], shell=False, stdout=subprocess.PIPE)

        sequence = process.communicate()[0].rstrip("\n").split("\n")[-1]
        #print contig, sequence
        if len(sequence) > 0:
            return sequence
        else:
            return None

    def get_islands(self, cutoff):
        "extracts islands below cutoff"

        islands = {}
        island_pattern = re.compile(r'(?P<contig>\d+) (?P<size>\d+) \d+\t;\t;\n')

        if cutoff > 0:
            for line in open(self.adj_islands, 'r'):
                m = island_pattern.search(line)

                if m and m.groupdict().has_key('contig'):
                    if int(m.groupdict()['size']) <= cutoff:
                        #print line, m.groupdict()['contig'], m.groupdict()['size']
                        islands[m.groupdict()['contig']] = True

        sys.stderr.write("islands(%d):%d\n" % (cutoff, len(islands.keys())))
        return islands
        
                
class Contig:

    def __init__(self, num, set=False, pet=False, length=0, coverage=0, k=None, children=None):
        self.num = num
        self.length = length
        self.coverage = coverage
        self.k = k
        self.children = children
        self.sequence = None

    def is_pet(self):
        if self.children:
            return True
        else:
            return False

    def fasta(self):
        headers = [str(self.num)]
        if self.length:
            headers.append(str(self.length))
        if self.coverage:
            headers.append(str(self.coverage))
        if self.children:
            headers.append(str(self.children))
            
        line = ">" + ' '.join(headers) + '\n'
        
        if self.sequence:
            seq = self.sequence
        else:
            seq = ""
        line += seq + "\n"

        return line

    def normalized_coverage(self):
        if self.coverage and int(self.coverage) >= 0 and self.length and int(self.length) > 1 and self.k:
            return "%.1f" % (float(self.coverage) / (float(self.length) - float(self.k) + 1.0))
        else:
            return None


if __name__ == '__main__':
    usage = "Usage: %prog library [options]"
    parser = OptionParser(usage=usage)
    parser.add_option("-d", "--dir", dest="dir", help="directory of assembly")
    parser.add_option("-k", "--k", dest="k", help="k")
    parser.add_option("-p", "--pet", dest="pet", help="size threshold for keeping paired-end contigs, default=0 (all kept)", type='int', default=0)
    parser.add_option("-s", "--set", dest="set", help="size threshold for keeping single end contigs, default=150", type='int', default=150)
    parser.add_option("-j", "--junction", dest="junction", help="size threshold for keeping single end contigs, default=kmer_size", type='int', default=0)
    parser.add_option("-f", "--fasta", dest="fasta", help="specific input fasta file")
    parser.add_option("-r", "--range", dest="range", help="size range", nargs=2, type='int')
    parser.add_option("-o", "--out", dest="out", help="output fasta")
    parser.add_option("-l", "--len", dest="len", help="contig length")
    parser.add_option("-c", "--contigs", dest="contigs", help="get contigs, e.g. 1/2/3")
    parser.add_option("-a", "--abyss", dest="abyss", help="ABySS version, e.g. 1.2.5")

    (options, args) = parser.parse_args()

    if len(args) != 1:
        parser.error("no library name given")

    elif not options.fasta and not options.dir:
        parser.error("no input fasta file given")
    elif not options.abyss:
        parser.error("no ABySS version given")
    else:
        if not (options.abyss in abyss_all):
            parser.error("ABySS version " + options.abyss + " not currently supported. Please use one of the following versions:" + abyss_all)

        #specific fasta file
        if options.fasta:
            assembly = Assembly(args[0], abyss=options.abyss, fasta=options.fasta, k=options.k)

            contigs = []
            if options.contigs:
                ids = options.contigs.split('/')
                contigs = assembly.get_contigs(length=options.len, len_range=options.range, sequence=True, ids=ids)
            else:
                contigs = assembly.get_contigs(length=options.len, len_range=options.range, sequence=True)

            if contigs and options.out:
                out = open(options.out, 'w')
                for contig in contigs:
                    out.write(contig.fasta())
                out.close()

        #directory given
        elif options.dir:
            outdir = None
            if options.out:
                outdir = os.path.split(options.out)[0]
            
            assembly = Assembly(args[0], abyss=options.abyss, loc=options.dir, k=options.k)

            sys.stderr.write("getting contigs\n")
            assembly.contigs = assembly.get_contigs(length=options.len, len_range=[options.pet,0], sequence=True)

            sys.stderr.write("getting indels\n")
            assembly.get_indels(outdir=outdir)

            contigs_remove = None
            junctions_remove = None
            if options.junction >= 0:
                sys.stderr.write("getting junctions\n")
                assembly.get_junctions(options.junction, outdir=outdir)
                sys.stderr.write("removing redundancy\n")
                contigs_remove, junctions_remove = assembly.remove_redundant()

            sys.stderr.write("getting islands\n")
            islands = assembly.get_islands(options.set)

            if options.out:
                out = open(options.out, 'w')
                
                for contig in assembly.contigs:
                    if contigs_remove and contigs_remove.has_key(contig.num):
                        continue
                    
                    #pe contigs
                    if contig.is_pet():
                        if int(contig.length) > options.pet:
                            out.write(contig.fasta())
                            
                    #se contigs
                    else:
                        if int(contig.length) > options.set:
                            out.write(contig.fasta())
                        elif int(contig.length) >= 2*int(assembly.k)-1:
                            if islands and islands.has_key(contig.num):
                                continue
                            out.write(contig.fasta())
                
                if options.junction >= 0:
                    for jn in assembly.junctions:
                        if junctions_remove and junctions_remove.has_key(jn.num):
                            continue
                        
                        out.write(jn.fasta())

                if assembly.indels:
                    for indel in assembly.indels:
                        out.write(indel.fasta())
                        
                out.close()
