#!/usr/bin/perl

use strict;

use Getopt::Long;
use Pod::Usage;
use vars qw(%OPT);
use Data::Dumper;

use Env qw(TRANSABYSS_PATH);
use TinyCfgReader;

GetOptions(\%OPT,
           "help|h",
           "man|m",
	   "align_contigs",
	   "xa2multi",
	   "fusions",
           "novelty",
           "gene_cover=s",
           "map=s",
           "annotation_cover=s",
	   "snv_indel",
           "start=s",
           "num=i",
           "lib=s",
           "cluster=s",
           "email=s",
           );

pod2usage(-verbose => 2) if $OPT{man};
pod2usage(1) if ($OPT{help});

my $package_path = $TRANSABYSS_PATH;
my $package_setup = "source $package_path/setup";
$package_setup =~ s/;*$//;

my $submitjobs = "$package_path/utilities/submitjobs.sh";
my $submitjobs_email = $OPT{email} ? $OPT{email} : TinyCfgReader::get_cluster_email();
my $cluster = $OPT{cluster} ? $OPT{cluster} : 0;
if ($cluster eq 'genesis') {
    $package_setup =~ s/projects/scratch/;
}

my $file = shift;

print "input:$file\n";

my $count = undef;

my @libs = split "\n", `cut -f 1,2,4 -d ' ' $file`;

for my $lib_loc(@libs) {
    my ($lib, $abyss_ver, $project) = split " ", $lib_loc;

    if ($OPT{lib}) {
	next if ($OPT{lib} ne $lib);
    }
    else {
	if ($OPT{start}) {
	    if (!defined $count && $OPT{start} eq $lib) {
	        $count = 1;
	    }
	}
	elsif (!defined $count) {
	    $count = 1;
	}

	last if $OPT{num} && $count > $OPT{num};
	next if (!defined $count);
    }	

    print "$lib\n";
    my $topdir = TinyCfgReader::get_project_info($project, "topdir");
    my $genome = TinyCfgReader::get_project_info($project, "reference");
    $submitjobs_email = TinyCfgReader::get_cluster_email($project);    
    my $verdir = "abyss-$abyss_ver";
    my $path = "$topdir/$lib/Assembly/$verdir";

    my $progress_log = "$topdir/$lib/log";

    if ($OPT{align_contigs}) {
	align_contigs($lib, $path, $genome, $project, $progress_log, "contigs");
    }
    
    if ($OPT{xa2multi})  {
        xa2multi($lib, $path, $project, $progress_log, $topdir, $abyss_ver);
    }
    
    if ($OPT{fusions}) {
	run_fusions($lib, $path, $topdir, $project, $genome, $progress_log);
    }
    
    if ($OPT{snv_indel}) {
	run_snv_indel($lib, $path, $topdir, $project, $genome, $progress_log);
    }
    
    if ($OPT{novelty}) {
	run_novelty($lib, $path, $topdir, $genome, $project, "contigs", $progress_log);
    }
        
    if ($OPT{gene_cover}) {
	run_coverage($lib, $path, $genome, $project, $progress_log);
    }
    
    $count ++;
}

sub align_contigs {
    my ($lib, $path, $genome, $project, $progress_log, $type) = @_;

    my $contigs = "$path/merge/$lib-$type.fa";

    my $job_name = "$lib-$type";
    my $out_dir = "$path/merge";
    my $cmd = "python $package_path/utilities/cluster_align.py";
    my $args = TinyCfgReader::get_args('cluster_align.py', $project);
    
    if (!-d "$out_dir/$job_name" && $OPT{cluster}) {
	my $mem = TinyCfgReader::get_cluster_mem('cluster_align.py', $project);
	my %hash = (
		    LIB => $lib,
		    GENOME => $genome,
		    CLUSTER => $cluster,
		    MEM => $mem,
		    TYPE => $type,
		    CONTIGS => $contigs,
		    OUT_DIR => $out_dir,
		    CONFIG_FILE => "$package_path/configs/transcriptome.cfg"
		    );
	$args = TinyCfgReader::replace_upper($args, \%hash);
	print "$cmd $args\n";
	`echo "------------------------------" >>$progress_log`;
	`echo "Running: $cmd $args" >>$progress_log`;
	#Run Python script to submit alignment jobs
	my @job_ids;
	`$package_setup`;
	open P, "$cmd $args |";
	while (<P>) {
	    my ($job_id) = $_ =~ /^Your job (\d+).+has been submitted/;
	    if ($job_id) {
		push @job_ids, $job_id;
	    }
	}
	close P;
    
	#filter alignments
	if (@job_ids) {
	    my $track_name = "\"$lib merged $type blat\"";
	    my $track = "$path/tracks/$lib.merge.$type.best.unique.m90.blat.psl";
	    my $align_dir = "$out_dir/cluster/$job_name/output";
        
	    my $args = TinyCfgReader::get_args('align_parser.py', $project);
	    my %hash = (
			BLAT_DIR => $align_dir,
			TRACK_NAME => $track_name,
			PSL => $track,
			CONTIGS => $contigs,
			GENOME => $genome,
			);
	    $args = TinyCfgReader::replace_upper($args, \%hash);
	    my $task = "time python $package_path/utilities/align_parser.py $args";
	    `echo "------------------------------" >>$progress_log`;
	    `echo "create track command: $task" >>$progress_log`;

	    my $clusterdir = "$path/tracks/cluster";
	    print "$clusterdir\n";
	    if (!-d $clusterdir) {
		`mkdir $clusterdir`;
	    }

	    my $jobfile = "$clusterdir/$lib-track-$type.sh";
	    my $cmd = "$package_setup;$task";
	    my $mem = TinyCfgReader::get_cluster_mem('align_parser.py', $project);
        
	    create_qsub_job($jobfile, [$cmd], {mem=>$mem, working_dir=>$clusterdir, email=>$submitjobs_email, hold_jid=>\@job_ids});
	    qsub_submit($jobfile, $OPT{cluster});
	}
	
    }
    `echo "------------------------------" >>$progress_log`;
}

sub xa2multi {  
    my ($lib, $path, $project, $progress_log, $topdir, $abyss_ver) = @_;
    
    #the amount memory configured for the cluster jobs
    my $mem = TinyCfgReader::get_cluster_mem('xa2multi', $project);
    
    #dirs
    my $anchordir = "$path/anchor";
    my $r2cdir = "$path/reads_to_contigs";
    my $clusterdir = "$r2cdir/cluster";
    if (!-d $clusterdir) {
        `mkdir -m 2775 $clusterdir`;
        `chmod a+w $clusterdir`;
    }    
    
    my $anchorbam = "$anchordir/$lib-anchor.bam";
    
    print `echo "------------------------------" >>$progress_log`;
    print `xa2multi ...`;
    
    my @cmds;
    push @cmds, $package_setup;
    push @cmds, "set -euo pipefail";    
    push @cmds, "samtools view -H $anchorbam > $r2cdir/$lib.bwa.xa.sam";
    push @cmds, "samtools view $anchorbam | xa2multi.pl >> $r2cdir/$lib.bwa.xa.sam";
    push @cmds, "samtools view -bhS $r2cdir/$lib.bwa.xa.sam -o $r2cdir/$lib.bwa.xa.bam";
    push @cmds, "samtools sort $r2cdir/$lib.bwa.xa.bam $r2cdir/$lib-contigs";
    push @cmds, "samtools index $r2cdir/$lib-contigs.bam";
    push @cmds, "rm $r2cdir/$lib.bwa.xa.sam $r2cdir/$lib.bwa.xa.bam";
    my $cmd = join ';', @cmds;    
    
    my $job = "$clusterdir/$lib-xa2multi.sh";
    open OUT, ">$job";
    print OUT "$cmd\n";
    close OUT;

    if ($cluster) {    
    	my $name = "$lib-xa2multi";
    	my $cur_args = TinyCfgReader::get_args('submitjobs.sh');
    	if ($submitjobs_email == -1) {
            $cur_args =~ s/ EMAIL//;
    	}
    	my %hash = (
    		    CLUSTER => $cluster,
    		    MQSUB => 'mqsub',
    		    CLUSTERDIR => $clusterdir,
    		    JOB => $job,
    		    NAME => $name,
    		    MEM => $mem,
    		    EMAIL => $submitjobs_email,
    		    );
    		    
    	$cur_args = TinyCfgReader::replace_upper($cur_args, \%hash);	    
    	print `$submitjobs $cur_args`;
    	`echo "Submitted: $submitjobs $cur_args" >>$progress_log`;
    }
    else {
	print "$cmd\n";
        `echo "Running: $cmd" >>$progress_log`;
	`$cmd`;
    }
    `echo "------------------------------" >>$progress_log`;
}


sub run_fusions {
    my ($lib, $path, $topdir, $project, $genome, $progress_log) = @_;
    
    my $script = "python $package_path/analysis/fusion.py";
    my $align_dir = "$path/merge/cluster/$lib-contigs/output";
    
    #get genome bam file
    opendir DIR, "$topdir/$lib/Reads_to_genome";
    my @genome_bam = grep /\.bam$/, readdir DIR;
    closedir DIR;

    if (-d $align_dir) {
	#run python script to generate batch jobs file
	my $args = TinyCfgReader::get_args('fusion.py', $project);
	my %hash = (
		    INPUT => $align_dir,
		    PATH => $path,
		    LIB => $lib,
		    TOPDIR => $topdir,
		    GENOME_BAM => $genome_bam[0],
		    GENOME => $genome,
		    );
	$args = TinyCfgReader::replace_upper($args, \%hash);
	
	`echo "------------------------------" >>$progress_log`;
	`echo "fusion command: $script $args" >>$progress_log`;
	
	`$package_setup;$script $args`;
			
	if ($OPT{cluster}) {
	    my $clusterdir = "$path/fusions/cluster";
	    my $batch_jobs = "$clusterdir/jobs";
	    my $mem = TinyCfgReader::get_cluster_mem('fusion.py', $project);
	    my @job_ids;
	    
	    #for each line in batch jobs file, create qsub job file and submit fusion job
	    if (-s $batch_jobs) {
		my $count = 1;
		open FILE, $batch_jobs;
		while (<FILE>) {
		    chomp;
		    my $cmd = $_;
		    my $jobfile = "$clusterdir/$lib-fusions-".$count++.".sh";
		    
		    create_qsub_job($jobfile, [$cmd], {mem=>$mem, working_dir=>$clusterdir, email=>$submitjobs_email});
		    my $job_id = qsub_submit($jobfile, $OPT{cluster});
		    if (defined $job_id) {
			push @job_ids, $job_id;
		    }
		}
		close FILE;
		
	    }
	    
	    #submit filtering job, with job dependencies on above fusion jobs
	    if (@job_ids) {
		my $args = TinyCfgReader::get_args('fusion.py_filter', $project);
		my %hash = (	
			    PATH => $path,
			    );
		$args = TinyCfgReader::replace_upper($args, \%hash);
		my $cmd = "$package_setup;$script $args";
		my $jobfile = "$clusterdir/$lib-fusion-filter.sh";
		
		create_qsub_job($jobfile, [$cmd], {mem=>$mem, working_dir=>$clusterdir, email=>$submitjobs_email, hold_jid=>\@job_ids});
		qsub_submit($jobfile, $OPT{cluster});
	    }
	}
	
    }
}

sub run_snv_indel {
    my ($lib, $path, $topdir, $project, $genome, $progress_log) = @_;
    
    my $script = "python $package_path/analysis/snv_caller.py";
    my $align_dir = "$path/merge/cluster/$lib-contigs/output";
    my $seq_dir = "$path/merge/cluster/$lib-contigs/input";
    
    #get genome bam file
    opendir DIR, "$topdir/$lib/Reads_to_genome";
    my @genome_bam = grep /\.bam$/, readdir DIR;
    closedir DIR;
    
    if (-d $align_dir && -d $seq_dir) {
	#run python script to generate batch jobs file
	my $args = TinyCfgReader::get_args('snv_caller.py', $project);
	my %hash = (
		    ALIGNS => $align_dir,
		    CONTIGS => $seq_dir,
		    PATH => $path,
		    LIB => $lib,
		    TOPDIR => $topdir,
		    GENOME_BAM => $genome_bam[0],
		    GENOME => $genome,
		    );
	$args = TinyCfgReader::replace_upper($args, \%hash);
	
	`echo "------------------------------" >>$progress_log`;
	`echo "snv_caller command: $script $args" >>$progress_log`;
	
	`$package_setup;$script $args`;
	
	if ($OPT{cluster}) {
	    my $clusterdir = "$path/snv/cluster";
	    my $batch_jobs = "$clusterdir/jobs";
	    my $mem = TinyCfgReader::get_cluster_mem('snv_caller.py', $project);
	    my @job_ids;
	    
	    #for each line in batch jobs file, create qsub job file and submit fusion job
	    if (-s $batch_jobs) {
		my $count = 1;
		open FILE, $batch_jobs;
		while (<FILE>) {
		    chomp;
		    my $cmd = $_;
		    my $jobfile = "$clusterdir/$lib-indel-".$count++.".sh";
		    
		    create_qsub_job($jobfile, [$cmd], {mem=>$mem, working_dir=>$clusterdir, email=>$submitjobs_email});
		    my $job_id = qsub_submit($jobfile, $OPT{cluster});
		    if (defined $job_id) {
			push @job_ids, $job_id;
		    }
		}
		close FILE;
		
	    }
	    
	    #submit filtering job, with job dependencies on above fusion jobs
	    if (@job_ids) {
		my $args = TinyCfgReader::get_args('snv_caller.py_filter', $project);
		my %hash = (	
			    PATH => $path,
			    );
		$args = TinyCfgReader::replace_upper($args, \%hash);
		my $cmd = "$package_setup;$script $args";
		my $jobfile = "$clusterdir/$lib-indel-filter.sh";
		my $mem = TinyCfgReader::get_cluster_mem('snv_caller.py_filter', $project);
		
		create_qsub_job($jobfile, [$cmd], {mem=>$mem, working_dir=>$clusterdir, email=>$submitjobs_email, hold_jid=>\@job_ids});
		qsub_submit($jobfile, $OPT{cluster});
	    }
	}
    }
}


sub run_novelty {
    my ($lib, $path, $topdir, $genome, $project, $type, $progress_log) = @_;

    my $outdir = "$path/novelty";
    if (!-d $outdir) {
	`mkdir $path/novelty`;
	`mkdir $outdir`;
    }

    my $script = "time python $package_path/analysis/model_matcher.py";
    my $track = "$path/tracks/$lib.merge.contigs.best.unique.m90.blat.psl";

    if (-s $track) {
	my $align_genome = TinyCfgReader::get_genome($genome);
	my $model_matcher_config = "$package_path/configs/model_matcher.cfg";
	my $args = TinyCfgReader::get_args('model_matcher.py', $project);
	my %hash = (
		    TRACK => $track,
		    PATH => $path,
		    OUTDIR => $outdir,
		    GENOME => $genome,
		    LIB => $lib,
		    );
	$args = TinyCfgReader::replace_upper($args, \%hash);
	my $task .= "$script $args";

	`echo "------------------------------" >>$progress_log`;
	`echo "novelty command: $task" >>$progress_log`;

	if (-d $outdir && -s $track) {
	    my $html = "time python $package_path/utilities/html_converter.py -i $path/novelty/$OPT{novelty}/events.txt -o $path/novelty/$OPT{novelty}/events.html -d $genome -b $path/novelty/$OPT{novelty}/events.bed -p $track";

	    my $clusterdir = "$path/novelty/cluster";
	    if (!-d $clusterdir) {
	        `mkdir $clusterdir`;
	    }

	    my $job = "$clusterdir/novelty.sh";
	    open OUT, ">$job";
	    print OUT "$package_setup;$task\n";
	    close OUT;
	    print "$job\n";

	    if ($OPT{cluster}) {
	        my $name = "$lib-novelty";
	        my $mem = TinyCfgReader::get_cluster_mem('model_matcher.py', $project);
	        $args = TinyCfgReader::get_args('submitjobs.sh', $project);
	        if ($submitjobs_email == -1) {
	            $args =~ s/ EMAIL//;
	        }
	        %hash = (
			CLUSTER => $cluster,
			MQSUB => 'mqsub',
			CLUSTERDIR => $clusterdir,
			JOB => $job,
			NAME => $name, 
			MEM => $mem,
			EMAIL => $submitjobs_email,
			);
		$args = TinyCfgReader::replace_upper($args, \%hash);
		print "$submitjobs $args\n";
		`$submitjobs $args`;
		`echo "Submitted: $submitjobs $args" >>$progress_log`;
	    }
	    else {
		print "$task\n";
		`echo "Running: $task" >>$progress_log`;
		`$task`;
		`echo "done" >>$progress_log`;
	    }
	`echo "------------------------------" >>$progress_log`;
	}
    }
}

sub run_coverage {
    my ($lib, $path, $genome, $project, $progress_log) = @_;

    my $bam = "$path/reads_to_contigs/$lib-contigs.bam";
    my $coverage = "$path/novelty/coverage.txt";
    my $track = "$path/tracks/$lib.merge.contigs.best.unique.m90.blat.psl";
    my $out = "$path/novelty/gene_coverage.txt"; #output in the same directory as coverage.txt
  
    if (-s $bam && -s $coverage && -s $track) {
	my $script = "time python $package_path/analysis/gene_coverage.py";
	my $opt;
	my $args = TinyCfgReader::get_args('gene_coverage.py', $project);
	my %hash = (
		    COVERAGE => $coverage,
		    R2C => $bam,
		    TRACK => $track,
		    LIB => $lib,
		    OUTPUT => $out,
		    );
	$args = TinyCfgReader::replace_upper($args, \%hash);
	my $task = "$script $args";
    
	`echo "------------------------------" >>$progress_log`;
	`echo "coverage command: $task" >>$progress_log`;
    
	my $clusterdir = "$path/novelty/cluster";
	print "$clusterdir\n";
	if (!-d $clusterdir) {
	    `mkdir $clusterdir`;
	}

	my $job = "$clusterdir/$lib"."_coverage.sh";
	open OUT, ">$job";
	print OUT "$package_setup;$task\n";
	close OUT;

	if ($cluster) {
	    my $name = "$lib-coverage";
	    my $mem = TinyCfgReader::get_cluster_mem('gene_coverage.py', $project);
	    $args = TinyCfgReader::get_args('submitjobs.sh', $project);
	    if ($submitjobs_email == -1) {
		$args =~ s/ EMAIL//;
	    }
	    %hash = (
		    CLUSTER => $cluster,
		    MQSUB => 'mqsub',
		    CLUSTERDIR => $clusterdir,
		    JOB => $job,
		    NAME => $name, 
		    MEM => $mem,
		    EMAIL => $submitjobs_email,
		    );
	    $args = TinyCfgReader::replace_upper($args, \%hash);
	    print"$submitjobs $args\n";
	    `$submitjobs $args`;
	    `echo "Submitted: $submitjobs $args" >>$progress_log`;
	}
	else
	{
	    print "$task\n";
	    `echo "Running: $task" >>$progress_log`;
	    `$task`;
	    `echo "done" >>$progress_log`;
	}
	`echo "------------------------------" >>$progress_log`;
    }
}

sub create_qsub_job {
    my $job_file = shift;
    my $cmds = shift;
    my $params = shift;
    
    my $queue = exists $params->{queue} ? $params->{queue} : 'all.q';
    my $mem = exists $params->{mem} ? $params->{mem} : '1G';
    
    open OUT, ">$job_file";
    print OUT "#! /bin/sh\n";
    print OUT "#\$ -S /bin/sh\n";
    print OUT "#\$ -R y\n";    
    print OUT "#\$ -q $queue\n";    
    print OUT "#\$ -l mem_token=$mem,mem_free=$mem,h_vmem=$mem\n";
    
    if (exists $params->{working_dir}) {
	print OUT "#\$ -wd $params->{working_dir}\n";
    }
    
    if (exists $params->{hold_jid}) {
	print OUT "#\$ -hold_jid ", join(',', @{$params->{hold_jid}}), "\n";
    }
    
    # email is sent when the job is aborted or rescheduled or suspended
    if ($params->{email} != -1) {
        print OUT "#\$ -M $params->{email}\n";
        print OUT "#\$ -m as\n";
    }
    
    print OUT "\n";
    for my $cmd(@$cmds) {
	print OUT "$cmd\n";
    }
    close OUT;
}

sub qsub_submit {
    my $job_file = shift;
    my $cluster = shift;
    
    print "submitting to $cluster: $job_file ... ";
    my $qsub_cmd = "ssh $cluster \"$package_setup;qsub $job_file\"";
    my $stdout = `$qsub_cmd`;
    my $job_id = undef;
    if ($stdout =~ /^Your job (\d+).+has been submitted/) {
        $job_id = $1;
    }
    print "$job_id\n";
    return $job_id;
}



__END__

=pod

=head1 NAME

analyze.pl - wrapper for analysis stage of Trans-ABySS

=head1 SYNOPSIS

analyze.pl <input_file> [options]

=head1 OPTIONS

=over 4

=item -cluster <CLUSTER_HEAD>

argument is the name of the cluster head node to submit jobs (if applicable). While small data can be processed locally on a desktop, it may be necessary to run jobs on cluster for large datasets.

=item -start <LIB_NAME>

start library, i.e. the name of the first library to be processed in the l
ist of libraries in the input file, can be used in combination with “-num”
 option (see below)

=item -num <NUM>

number of libraries to process from "start library". Use this option in co
mbination with “-start” option to specify the libraries that need to be pr
ocessed. If only “-start” option is specified without “-num” option, then 
all libraries from the start library onwards will be processed.

=item -lib <LIB_NAME>

library that needs to be processed (for processing single library). If onl
y a single library needs to be processed, use this option instead of “-sta
rt” and “-num” combination.

Running “analyze.pl” without “-start”, “-num”, or “-lib” options will proces
s all libraries in the input file.

=item -fusions

run fusion script, will generate output as two text files under "fusions" directory: a "<REF>_fusions.txt" and a "<REF>_fusions_filtered.txt", e.g. hg18_fusions.txt and hg18_fusions_filtered.txt

=item -track

generates filtered track file from contig-to-genome Blat alignments

=item -novelty <OUT_DIR>

find novelty, argument is the name of output directory (inside "novelty" folder), e.g. blat. This can also generate coverage statistics ("coverage.txt" and "gene2transcript.map") for calculating gene expression level (see below).

=item -gene_cover <OUT_FILE>

determine gene expression, using the "coverage.txt" generated by "-novelty" option that maps contigs to transcript annotation, argument is the output file name (will be generated under novelty/blat folder)

=item -map <GENE_TO_TRANSCRIPT_MAPPING_FILE>

optional mapping file for gene expression calculation, for cases when there are missing gene names in the "coverage.txt" file, due to missing gene names in annotation files. If this file is not specified, coverage for transcripts without gene names will be ignored.

=item -annotation_cover <OUT_FILE>

determine the coverage of annotated genes or transcripts, using "coverage.txt" generated by "-novelty" option, argument is the output file name (will be generated under novelty/blat folder). If the annotation used by "-novelty" option is transcripts, this will return the transcript coverage; if the annotation used by "-novelty" option is genes, this will return the gene coverage.

=item -snv_indel

detect SNVs and indels, using the contigs_to_reference alignment track file and reads_to_contigs alignment file.

=item -fusions_filter

filter fusion results (gene_fusions_csv.txt and lsr_csv.txt) to generate filtered results (gene_fusions2_csv.txt and lsr2_csv.txt)

=back

=head1 DESCRIPTION

This wrapper script can execute the followings after alignments are finished: 1)find fusions; 2)generate filtered track; 3)find novel events; 4)calculate gene expression level

=head1 AUTHOR

Readman Chiu rchiu@bcgsc.ca

=head1 EXAMPLE

./analyze.pl /home/mapper/trans-abyss/gsc/pipeline/libs/100823 -start HS0669 -num 2 -fusions -cluster apollo

./analyze.pl libs/HS0382 -track -cluster apollo

./analyze.pl input/sample -novelty blat -cluster apollo

./analyze.pl input/sample -gene_cover gene_expression.txt

./analyze.pl input/sample -snv_indel -cluster apollo


=cut
