#!/usr/bin/perl

use Data::Dumper;
use Getopt::Long;
use Pod::Usage;
use vars qw(%OPT);
use strict;
use Cwd;

use Env qw(TRANSABYSS_PATH ANCHOR_DIR);
use TinyCfgReader;

GetOptions(\%OPT,
           "help|h",
           "man|m",
           "start=s",
           "num=i",
           "lib=s",
           "cluster=s",
	   "make_dir",
	   "get_reads",
	   "copy_bam",
	   "fem",
	   "anchor_finalize",
	   "genesis_dir=s",
           "email=s",
	   "copy_dir",
           );

pod2usage(-verbose => 2) if $OPT{man};
pod2usage(1) if ($OPT{help});

my $curdir = cwd;
print "$curdir\n";

my $package_path = $TRANSABYSS_PATH;
my $conf_path = $package_path . '/configs';
my $util_path = $package_path . '/utilities';
my $wrapper_path = $package_path . '/wrappers';

my $package_setup = "source $package_path/setup";
$package_setup =~ s/;*$//;

my $submitjobs = "$util_path/submitjobs.sh";
my $submitjobs_email = $OPT{email} ? $OPT{email} : TinyCfgReader::get_cluster_email();

my $file = shift;
my $count = undef;

my $cluster = $OPT{cluster} ? $OPT{cluster} : 0;
if ($cluster eq 'genesis') {
    $package_setup =~ s/projects/scratch/;
}

my $make_dir = $OPT{make_dir} ? $OPT{make_dir} : 0;
my $get_reads = $OPT{get_reads} ? $OPT{get_reads} : 0;
my $fem = $OPT{fem} ? 1 : 0;
my $finalize = $OPT{anchor_finalize} ? 1 : 0;
my $copy_bam = $OPT{copy_bam} ? 1 : 0;
my $genesis_dir = $OPT{genesis_dir} ? "/genesis/scratch/$OPT{genesis_dir}" : undef;

open FILE, $file;
while (<FILE>) {
  chomp;  
  next unless $_;
#  my ($lib, $ver, $source, $project, $read_length) = split ' ', $_;
#  $read_length = 50 unless $read_length; #set a default length
  my ($lib, $ver, $source, $project) = split ' ', $_;

  $source =~ s/\/+$//;
  my ($source_lib) = $source =~ /.+\/(\S+)$/;

  if ($OPT{lib}) {
      next if ($OPT{lib} ne $lib);
  }
  else {
      if ($OPT{start}) {
    	  if (!defined $count && $OPT{start} eq $lib) {
    	      $count = 1;
    	  }
      }
      elsif (!defined $count) {
    	  $count = 1;
      }
      last if $OPT{num} && $count > $OPT{num};
      next if (!defined $count);
  }

  my $project_dir = TinyCfgReader::get_project_info($project, "topdir");
  my $topdir = TinyCfgReader::get_project_info($project, "topdir") . "/$lib";
  my $genome = TinyCfgReader::get_project_info($project, "reference");
  $submitjobs_email = TinyCfgReader::get_cluster_email($project);

  print "topdir: $topdir; reference: $genome\n";

  my $progress_log = "$topdir/log";

  if ($make_dir) {
    setup($lib, $topdir, $ver, $source, $genome, $progress_log);
  }

  if ($get_reads) {
      get_reads($lib, $topdir, $ver, $source, $project, $progress_log);
  }

  if ($copy_bam) {
      copy_bam($lib, $topdir, $genome, $project, $progress_log);
  }
  
  if ($fem) {      
      fem($lib, $topdir, $ver, $source, $project, $source_lib, $progress_log);
  }  

  if ($finalize) {
      finalize($lib, $topdir, $ver, $project, $progress_log);
  }

  if ($OPT{copy_dir} && defined $genesis_dir) {
      copy_dir($project, $lib, $genesis_dir, $project_dir);
  }

  $count++;
}
close FILE;

chdir($curdir);

sub get_reads {
    my ($lib, $topdir, $ver, $source, $project, $progress_log) = @_;

    my $asmdir = "$topdir/Assembly";
    my $verdir = "$asmdir/abyss-$ver";

    `echo "------------------------------" >>$progress_log`;
#    `echo "Setup directories ... " >>$progress_log`;
       
    #Anchor Python scripts
    my $anchor_setup_script = "$ANCHOR_DIR/lib/anchor_setup.py";
    
    #check whether files exist
    unless (-e $anchor_setup_script){
        print "ERROR: Cannot find $anchor_setup_script\n";
        return;
    }
    
    #existing files
    my $merged_contigs_fa_file = "$verdir/merge/$lib-contigs.fa"; #will be created in stage 0
    my $input_reads_file = "$verdir/source/$lib.in"; #text file
    
    #check whether files exist
    unless (-e $input_reads_file){
        print "ERROR: Cannot find $input_reads_file\n";
        return;
    }
    
    #check whether input reads are bam files
    my $reads_in_bams = 0; #false
    my $converted = 0; #false
    #read the input reads text file
    open FILE_READS, $input_reads_file;
    while (<FILE_READS>) {
        chomp;
        if ($_ =~ /\.bam$/){
            $reads_in_bams = 1;
        }
        elsif ($_ =~ /\.fq\.gz$/){
            $converted = 1;
        }
        elsif ($_ =~ /\.fastq\.gz$/){
            $converted = 1;
        }                       
    }
    close FILE_READS;
    
    #directories to be created
    my $working_dir = "$verdir";
    my $anchor_working_dir = "$working_dir/anchor";
    my $anchor_reads_dir = "$anchor_working_dir/reads";      
    
    print "Set up the directories and convert reads for ANCHOR ...\n";
    
    # Set up the directories and convert the reads to the appropriate format if needed.
    if ($reads_in_bams || $converted){
        print "No conversion/compression needed.\n";
    	my $anchor_setup_cmd = "mkdir $anchor_working_dir && ln -s $input_reads_file $anchor_working_dir/$lib.in && ln -s $merged_contigs_fa_file $anchor_working_dir/$lib-contigs.fa";
    	
    	print `echo "Set up directories for ANCHOR:\n$anchor_setup_cmd" >>$progress_log`;
    	my $setup_result = `$anchor_setup_cmd`;
    	print $setup_result; 
    }
    else{
        print "Reads will be converted to compressed FASTQ.\n";
        #command to run the anchor setup script (to set up the dirs and reads)
        my $cmd = "$package_setup && cd $working_dir && python $anchor_setup_script $merged_contigs_fa_file $input_reads_file $anchor_working_dir $anchor_reads_dir";
        
        #check whether the cluster head node was specified
        if ($cluster) {
            #ssh to cluster head node and run the command        
            my $anchor_setup_cmd = "ssh $cluster '$cmd' 2>&1";
                    
            print `echo "Set up directories for ANCHOR and convert reads to FASTQ:\n$anchor_setup_cmd" >>$progress_log`;
            my $setup_result = `$anchor_setup_cmd\n`;
            print $setup_result;
        }
        else{
            print "Cluster head node was not specified. Run locally...\n";
            print "$cmd\n";
            `echo "Running: $cmd" >>$progress_log`;
            `$cmd`;
        }
    }
    # endif
    ### end

    #`echo "getting reads done" >>$progress_log`;
    `echo "------------------------------" >>$progress_log`;
}

sub copy_bam {
    my ($lib, $topdir, $genome, $project, $progress_log) = @_;
    print "Please put your code in:\n $package_path/setup.pl (sub copy_bam)\n for copying JAGUAR's BAM file to the \"Reads_to_genome\" directory!\n";
    
    ############################################################################################
    # TODO: Put your code here for copying JAGUAR's BAM file to the "Reads_to_genome" directory.
    ############################################################################################
    
}

sub setup {
  my ($lib, $topdir, $ver, $source, $genome, $progress_log) = @_;

  `echo "------------------------------" >>$progress_log`;
  `echo "Setup directories ... " >>$progress_log`;

  my @cmds;

  if (!-d $topdir) {
    push @cmds, "mkdir -m 2775 $topdir";
    push @cmds, "chgrp trans-abyss $topdir";
  }

  my $asmdir = "$topdir/Assembly";
  my $readsdir = "$topdir/Reads_to_genome";
  my $verdir = "$asmdir/abyss-$ver";
  my $current = "$asmdir/current";

  if (!-d $asmdir) {
    push @cmds, "mkdir -m 2775 $asmdir";
  }
  if (!-d $readsdir) {
    push @cmds, "mkdir -m 2775 $readsdir";
  }
  if (!-d $verdir) {
    push @cmds, "mkdir -m 2775 $verdir";
  }
  if (!-l $current) {
    push @cmds, "ln -s abyss-$ver $current"; #symlink points to relative path!
  } else {
      #if ("$asmdir/$target" ne $verdir) {
      push @cmds, "rm $current";
      push @cmds, "ln -s abyss-$ver $current";
    #}
  }

  if (-d $source) {
    my $source_link = "$verdir/source";
    if (!-l $source_link) {
        push @cmds, "ln -s $source $source_link";
    }

    opendir DIR, $source;
    my @ks = sort {$a<=>$b} map {/(\d+)/} grep /^k(\d+)$/, readdir DIR;
    closedir DIR;

    for my $k(sort @ks) {
        if (-d "$source/k$k" && !-d "$verdir/k$k") {
            push @cmds, "mkdir -m 2775 $verdir/k$k";
        }
    }
  }

  #merge
  my $mergedir = "$verdir/merge";
  #my $mergingdir = "$mergedir/merging";
  my $clusterdir = "$mergedir/cluster";
  #my $alignsdir = "$mergedir/aligns";

  push @cmds, "mkdir -m 2775 $mergedir" if !-d $mergedir;
  #push @cmds, "mkdir -m 2775 $mergingdir" if !-d $mergingdir;
  if (!-d $clusterdir) {
    push @cmds, "mkdir -m 2775 $clusterdir";
    push @cmds, "chmod a+w $clusterdir";
  }
  #if (!-d $alignsdir) {
  #  push @cmds, "mkdir -m 2775 $alignsdir";
  #  push @cmds, "mkdir -m 2775 $alignsdir/$genome"; #hg18";
  #  push @cmds, "mkdir -m 2775 $alignsdir/$genome/blat_34"; #hg18/blat_34";
  #}
  
#  my $scaffolddir = "$verdir/scaffold";
#  push @cmds, "mkdir -m 2775 $scaffolddir" if !-d $mergedir;

  #dirs
  my $fusionsdir = "$verdir/fusions";
  my $tracks = "$verdir/tracks";
  my $r2c = "$verdir/reads_to_contigs";
  my $variants = "$verdir/novelty";
#  my $anchor = "$verdir/anchor";
  my $snv = "$verdir/snv";

  for my $d($fusionsdir, $tracks, $r2c, $variants, $snv) {
    push @cmds, "mkdir -m 2775 $d" if !-d $d;
  }

  for my $cmd(@cmds) {
    print "$cmd\n";
    `$cmd`;
  }

  `echo "setup directories done" >>$progress_log`;
  `echo "------------------------------" >>$progress_log`;
}

sub fem{
    my ($lib, $topdir, $ver, $source, $project, $source_lib, $progress_log) = @_;
  
    print `echo "------------------------------" >>$progress_log`;
    print `echo "Filter,Extend,Merge assemblies ..." >>$progress_log`;
    
    #the amount memory configured for the cluster jobs
    my $mem = TinyCfgReader::get_cluster_mem('fem', $project);
    
    #wrapper scripts needed 
    my $filter_extend_script = "$wrapper_path/abyss-ta-filter";
    my $merge_script = "$wrapper_path/abyss-rmdups-iterative";
  
    #check whether wrapper scripts exist
    unless (-e $filter_extend_script){
        print "ERROR: Cannot find $filter_extend_script";
        return;
    }
    unless (-e $merge_script){
        print "ERROR: Cannot find $merge_script";
        return;
    }
    
    #dirs
    my $asmdir = "$topdir/Assembly";
    my $verdir = "$asmdir/abyss-$ver";
    my $working_dir = "$verdir/merge/cluster";
    
    unless (-d $working_dir){
        print "ERROR: Cannot find $working_dir";
        return;
    }
    
    #job scripts to be created    
    my $fem_job_script = "$working_dir/$lib-fem.sh";
    
    ### Write the script for FEM
    open OUT, ">$fem_job_script";
    print OUT "\#! /bin/sh\n";
    
    if ($cluster) { 
        print OUT "\#\$ -S /bin/sh\n";
        print OUT "\#\$ -R y\n";    
        print OUT "\#\$ -wd $working_dir\n";
        print OUT "\#\$ -l mem_token=$mem,mem_free=$mem,h_vmem=$mem\n";
        print OUT "\#\$ -pe ncpus 8\n"; # use 8 cpus
        if ($submitjobs_email != -1) {
            # email is sent when the job is aborted or rescheduled or suspended
            print OUT "\#\$ -M $submitjobs_email\n";
            print OUT "\#\$ -m as\n";
        }        
    }  
    
    print OUT "\n";
    print OUT "$package_setup\n";
    print OUT "echo \$(hostname)\n";    
    print OUT "set -eu\n";    
    print OUT "cd $verdir\n";
    
    opendir DIR, $verdir;
    my @ks = sort {$a<=>$b} map {/(\d+)/} grep /^k(\d+)$/, readdir DIR;
    closedir DIR;

    #filter, extend each k
    @ks = sort @ks;
    for my $k(@ks) {
        if (-d "$verdir/k$k") {
            print OUT "time $filter_extend_script $source $k $lib k$k 151\n";
        }
    }    
    
    #merge all k
    print OUT "$merge_script $verdir $verdir/merge $lib\n";
    
    close OUT;    
    
    #check whether the cluster head node was specified
    if ($cluster) {    
    	#submit the script to cluster
    	my $fem_cmd = "ssh $cluster qsub $fem_job_script 2>&1";
    	print `echo "Filter,extend,merge:\n$fem_cmd" >>$progress_log`;        
    	my $fem_result = `$fem_cmd`; #eg. Your job 1454331 ("A04201-fem.sh") has been submitted
    	print $fem_result;
    }
    else{
        my $fem_cmd = "sh $fem_job_script";
        print "Cluster head node was not specified. Run locally...\n";
        print "$fem_cmd\n";
        `echo "Running: $fem_cmd" >>$progress_log`;
        `$fem_cmd`;        
    }
    ### end

    print `echo "------------------------------" >>$progress_log`;    
}

sub finalize {
    my ($lib, $topdir, $ver, $project, $progress_log) = @_;
    
    #the amount memory configured for the cluster jobs
    my $mem = TinyCfgReader::get_cluster_mem('anchor_pipeline.py', $project);
    
    #Anchor Python script
    my $anchor_pipeline_script = "$ANCHOR_DIR/anchor_pipeline.py";
    
    #check whether files exist
    unless (-e $anchor_pipeline_script){
        print "ERROR: Cannot find $anchor_pipeline_script";
        return;
    }    

    #dirs
    my $asmdir = "$topdir/Assembly";
    my $verdir = "$asmdir/abyss-$ver";    
    my $working_dir = "$verdir/anchor";
    unless (-d $working_dir){
        print "ERROR: Cannot find $working_dir";
        return;
    }    
    
    #files
    my $merged_contigs_fa_file = "$working_dir/$lib-contigs.fa";
    my $input_reads_file = "$working_dir/$lib.in"; #text file
    unless (-e $merged_contigs_fa_file){
        print "ERROR: Cannot find $merged_contigs_fa_file";
        return;
    }
    unless (-e $input_reads_file){
        print "ERROR: Cannot find $input_reads_file";
        return;
    }
    
    #check whether input reads are bam files
    my $reads_in_bams = 0; #false
    #read the input reads text file
    open FILE_FINALIZE, $input_reads_file;
    while (<FILE_FINALIZE>) {
        chomp;
        if ($_ =~ /\.bam$/){
            $reads_in_bams = 1;
        }
    }
    close FILE_FINALIZE;
    
    #job scripts to be created
    my $finalize_job_script = "$working_dir/$lib-finalize.sh";
       
    print `echo "------------------------------" >>$progress_log`;
    print `Finalize the merged assembly with ANCHOR...`;
    
    my $use_cluster = '';
    
    ###  finalize the assembly with ANCHOR.
    #script for running the finalizing job
    open OUT, ">$finalize_job_script";
    print OUT "\#! /bin/sh\n";
    if ($cluster) {
        print OUT "\#\$ -S /bin/sh\n";
        print OUT "\#\$ -R y\n";    
        print OUT "\#\$ -wd $working_dir\n";
        print OUT "\#\$ -l mem_token=$mem,mem_free=$mem,h_vmem=$mem\n";
        if ($submitjobs_email != -1) {
            # email is sent when the job is aborted or rescheduled or suspended
            print OUT "\#\$ -M $submitjobs_email\n";
            print OUT "\#\$ -m as\n";
        }
        $use_cluster = "-c $cluster";
    }
    
    my $bam_option = '';
    if ($reads_in_bams){
        $bam_option = "-b"; #b: reads are stored in BAM files
    }    
    
    print OUT "\n";
    print OUT "$package_setup\n";
    print OUT "echo \$(hostname)\n";
    print OUT "cd $working_dir\n";
    print OUT "python $anchor_pipeline_script -r $bam_option $use_cluster -F $lib-contigs.fa $lib.in $lib\n";
    print OUT "if [ ! -e $lib.final.fa ]; then\n";
    print OUT "echo \"ERROR: Cannot find $lib.final.fa\"\n";    
    print OUT "exit 1\n";
    print OUT "fi\n";
#    print OUT "ln -s $lib.final.fa $lib-anchor.raw.fa\n";    
    print OUT "if [ \$(ls *.sorted.[0-9][0-9][0-9][0-9].bam 2> /dev/null | wc -l) -gt 0 ]; then\n";
    print OUT "echo \"ERROR: One or more BAM files might not have been sorted.\"\n";
    print OUT "exit 1\n";
    print OUT "fi\n";
    print OUT "if [ ! -e $lib.bam_files.in ]; then\n";
    print OUT "echo \"ERROR: Cannot find $lib.bam_files.in\"\n";
    print OUT "exit 1\n";
    print OUT "fi\n";        
    print OUT "numbam=\$(cat $lib.bam_files.in | awk 'NF = 1' | awk 'END{print NR}')\n";
    print OUT "bamfiles=\$(awk 'BEGIN { RS = \"\"; OFS = \" \"} {\$1 = \$1; print }' $lib.bam_files.in)\n";    
    print OUT "if [ \$numbam -gt 1 ]; then\n";
    print OUT "samtools merge $lib-anchor.bam \$bamfiles && samtools index $lib-anchor.bam\n";
    print OUT "elif [ \$numbam -eq 1 ]; then\n"; 
    print OUT "ln -s \$bamfiles $lib-anchor.bam\n";
    print OUT "ln -s \$bamfiles.bai $lib-anchor.bam.bai\n";
    print OUT "else\n";
    print OUT "echo \"ERROR: Cannot find any BAM files!\"\n";
    print OUT "fi\n";
    close OUT;
    #script EOF
        
    if ($cluster) {    
        #submit the script to cluster
        my $anchor_pipeline_cmd = "ssh $cluster qsub $finalize_job_script 2>&1";
        print `echo "finalize the assembly with ANCHOR:\n$anchor_pipeline_cmd" >>$progress_log`;        
        my $anchor_result = `$anchor_pipeline_cmd`; #eg. Your job 1454331 ("A04201-finalize.sh") has been submitted
        print $anchor_result;
    }
    else {
        print "Cluster head node was not specified. Run locally...\n";
        my $fin_cmd = "sh $finalize_job_script";        
        print "$fin_cmd\n";
        `echo "Running: $fin_cmd" >>$progress_log`;
        `$fin_cmd`;
    }    
    ### end

    print `echo "------------------------------" >>$progress_log`;
}

sub copy_dir {
    my ($project, $lib, $genesis_dir, $topdir) = @_;

    my $cmd = "time rsync -aHu --stats --exclude=*.fq $genesis_dir/$project/$lib $topdir";

    print "$cmd\n";

    if ($cluster) {
        if (! -d "$topdir/cluster") {
            `mkdir -m 2775 $topdir/cluster`;
            `chgrp trans-abyss $topdir/cluster`;
        }
        my $jobs_file = "$topdir/cluster/$lib-copy.sh";
        open OUT, ">$jobs_file";
        print OUT "$cmd\n";
        close OUT;

        my $mem = '1G';
        my $args = TinyCfgReader::get_args('submitjobs.sh', $project);
        if ($submitjobs_email == -1) {
            $args =~ s/ EMAIL//;
        }
        my %hash = (
                    CLUSTER => $cluster,
                    MQSUB => 'mqsub',
                    CLUSTERDIR => "$topdir/cluster",
                    JOB => $jobs_file,
                    NAME => "$lib-copy",
                    MEM => $mem,
                    EMAIL => $submitjobs_email,
                    );
        my $args1 = TinyCfgReader::replace_upper($args, \%hash);
        $args1 =~ s/all.q/thosts.q/;
        print `$submitjobs $args1`;

    } 
    else {
        print "$cmd\n";
    }
}


__END__

=pod

=head1 NAME

setup.pl - wrapper for processing and merging Trans-ABySS assemblies

=head1 SYNOPSIS

setup.pl <input_file> [options]

=head1 OPTIONS

=over 4

=item -cluster <CLUSTER_HEAD>

name of the cluster head node to submit jobs (if applicable). While small data can be processed locally on a desktop, it may be necessary to run jobs on cluster for large datasets.

=item -start <LIB_NAME>

start library, i.e. the name of the first library to be processed in the list of libraries in the input file, can be used in combination with “-num” option (see below)

=item -num <NUM>

number of libraries to process from "start library". Use this option in combination with “-start” option to specify the libraries that need to be processed. If only “-start” option is specified without “-num” option, then all libraries from the start library onwards will be processed.

=item -lib <LIB_NAME>

library that needs to be processed (for processing single library). If only a single library needs to be processed, use this option instead of “-start” and “-num” combination.

Running “setup.pl” without “-start”, “-num”, or “-lib” options will process all libraries in the input file.

=item -make_dir

just creates working directory structure

=item -make_contigs

just creates contigs for all k-assemblies (working directory must have been created)

=item -get_reads

gets reads for the assembly

=item -copy_bam

copy genome bam file over

=item -merge

just merge contigs from all k-assemblies into a final assembly (contigs for all k-assemblies must have been created)

=item -scrub

send scrubbing job to genesis (must provide -genesis_dir)

=item -copy_scrub

copy scrubbing results back (must provide -genesis_dir)

=item -genesis_dir

genesis topdir where library is located (e.g. transcriptome2, validations, FL_BAC)

=back

=head1 DESCRIPTION

This wrapper script sets up the directory structure for Trans-ABySS, creates contigs files for each k-assembly, and submit a job to cluster for merging all-k assemblies into a final assembly.  By default, it will try to carry out all 3 steps. It is possible to run only a specific step using the '-make_dir', '-make_contigs', and '-merge' option.

=head1 AUTHOR

Readman Chiu rchiu@bcgsc.ca

=head1 EXAMPLE

setup.pl input/sample

setup.pl /home/mapper/trans-abyss/gsc/pipeline/libs/100823 -lib HS0669

setup.pl /home/mapper/trans-abyss/gsc/pipeline/libs/100823 -start HS0669

setup.pl /home/mapper/trans-abyss/gsc/pipeline/libs/100823 -start HS0669 -num 2 -cluster apollo

=cut
